# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
import re
import sys
import inspect
import warnings
from copy import copy
from collections import defaultdict
from collections import namedtuple
from collections.abc import Collection
from collections.abc import MutableSequence
from collections.abc import Sequence
from importlib import import_module
from functools import partial
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Union
from weakref import WeakKeyDictionary
import typing as t

from fresco.exceptions import ResponseException
from fresco.response import Response
from fresco.request import Request
from fresco.requestcontext import context
from fresco.routeargs import RouteArg
from fresco.types import WSGIApplication
from fresco.types import ViewCallable
from fresco.util.cache import make_cache
from fresco.util.common import fq_path
from fresco.util.urls import join_path
from fresco.util.object import classorinstancemethod

RFC2518_METHODS = PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK = (
    "PROPFIND",
    "PROPPATCH",
    "MKCOL",
    "COPY",
    "MOVE",
    "LOCK",
    "UNLOCK",
)

RFC2616_METHODS = GET, HEAD, POST, PUT, DELETE, OPTIONS, TRACE, CONNECT = (
    "GET",
    "HEAD",
    "POST",
    "PUT",
    "DELETE",
    "OPTIONS",
    "TRACE",
    "CONNECT",
)

RFC3253_METHODS = (
    VERSION_CONTROL,
    REPORT,
    CHECKOUT,
    CHECKIN,
    UNCHECKOUT,
    MKWORKSPACE,
    UPDATE,
    LABEL,
    MERGE,
    BASELINE_CONTROL,
    MKACTIVITY,
) = (
    "VERSION-CONTROL",
    "REPORT",
    "CHECKOUT",
    "CHECKIN",
    "UNCHECKOUT",
    "MKWORKSPACE",
    "UPDATE",
    "LABEL",
    "MERGE",
    "BASELINE-CONTROL",
    "MKACTIVITY",
)

RFC3648_METHODS = (ORDERPATCH,) = ("ORDERPATCH",)

RFC3744_METHODS = (ACL,) = ("ACL",)

RFC5323_METHODS = (SEARCH,) = ("SEARCH",)

RFC5789_METHODS = (PATCH,) = ("PATCH",)

ALL_METHODS = HTTP_METHODS = set(
    RFC2518_METHODS
    + RFC2616_METHODS
    + RFC3253_METHODS
    + RFC3648_METHODS
    + RFC5323_METHODS
    + RFC5789_METHODS
)

#: Shortcut for specifying all methods as a kwarg
ALL = "ALL"

__all__ = [
    "ALL_METHODS",
    "Pattern",
    "Route",
    "DelegateRoute",
    "RouteCollection",
    "routefor",
]
__all__ += [method.replace("-", "_") for method in ALL_METHODS]


#: Encapsulate a pattern match on a path
PathMatch = namedtuple(
    "PathMatch", ["path_matched", "path_remaining", "args", "kwargs"]
)


class RouteTraversal(
    namedtuple("RouteTraversal", "route view args kwargs collections_traversed")
):
    """
    Encapsulate a route traversal.

    Each ``RouteTraversal`` object contains:

    - ``route`` the final route traversed, allowing access to the view
        associated with the path.
    - ``view``, the resolved view callable
    - ``args`` - positional args to be passed to the view. This is a
        combination of args extracted from the path and any added when
        constructing the route.
    - ``kwargs`` - keyword args to be passed to the view. This is a
        combination of kwargs extracted from the path and any added when
        constructing the route.
    -  ``collections_traversed`` - a list of ``TraversedCollection``
        objects, containing information on which RouteCollections were
        encountered during traversal, the route selected at each stage
        and any args/kwargs associated with that phase of the traversal.
    """

    def replace(self, viewspec, traversal_kwargs=None, **kwargs):
        """
        Return a new ``RouteTraversal`` with the traversal_kwargs or other
        fields replaced for the route identified by ``viewspec``.

        Example:

            >>> from fresco import FrescoApp
            >>> app = FrescoApp()
            >>> @route('/<lang:str>/index.html', GET, homepage, name='home')
            ... def homepage(lang):
            ...     return Response({'fr': 'Bonjour!', 'en': 'Hello!'}[lang])
            ...
            >>> traversal = next(app.get_route_traversals('/fr/index.html'))
            >>> en_traversal = traversal.replace('home', {'lang': 'en'})
            >>> en_traversal.build_path()
            '/en/index.html'

        """
        if traversal_kwargs is not None:
            kwargs["traversal_kwargs"] = traversal_kwargs
        if isinstance(viewspec, str):
            viewspecs = viewspec.split(":")
        else:
            viewspecs = [viewspec]
        collections_traversed_iter = iter(self.collections_traversed)
        route = None
        for item in viewspecs:
            while True:
                ct = next(collections_traversed_iter, None)
                if ct is None:
                    raise RouteNotFound(viewspec)
                try:
                    route = ct.collection.routefor(item)
                    break
                except RouteNotFound:
                    continue

        new_collections_traversed = [
            c._replace(**kwargs) if (c.route is route) else c
            for c in self.collections_traversed
        ]
        return self._replace(collections_traversed=new_collections_traversed)

    def build_path(self):
        """
        Reconstruct the path from a route traversal
        """
        paths = []
        for c in self.collections_traversed:
            path, _, _ = c.route.path(*c.traversal_args, **c.traversal_kwargs)
            paths.append(path)
        return "".join(paths)


#: An item of RouteTraversal.collections_traversed
TraversedCollection = namedtuple(
    "TraversedCollection",
    "collection path route args kwargs traversal_args traversal_kwargs",
)


_marker = object()


class RouteNotReady(RuntimeError):
    """
    getview was called on a route mapping to an instance method, but the
    class has yet been instantiated
    """


class URLGenerationError(Exception):
    """\
    Was not possible to generate the requested URL.
    """


class RouteNotFound(Exception):
    """\
    The named route does not exist in the RouteCollection.
    """


class Pattern(object):
    """\
    Patterns are matchable against URL paths using their ``match`` method. If a
    path matches, this should return a tuple of ``(positional_arguments,
    keyword_arguments)`` extracted from the URL path. Otherwise this method
    should return ``None``.

    Pattern objects may also be able to take a tuple of
    ``(positional_arguments, keyword_arguments)`` and return a corresponding
    URL path.
    """

    segments: list["PatternSegment"]

    def match(self, path):
        """
        Should return a tuple of ``(positional_arguments, keyword_arguments)``
        if the pattern matches the given URL path, or None if it does not
        match.
        """
        raise NotImplementedError

    def pathfor(self, *args, **kwargs):
        """
        The inverse of ``match``, this should return a URL path
        for the given positional and keyword arguments, along with any unused
        arguments.

        :return: a tuple of ``(path, remaining_args, remaining_kwargs)``
        """
        raise NotImplementedError()

    def path_argument_info(self):
        """
        Return information about the arguments required for ``pathfor``
        """
        raise NotImplementedError()

    def add_prefix(self, prefix):
        """
        Return a copy of the pattern with the given string prepended
        """
        raise NotImplementedError()


class Converter:
    """
    Responsible for converting arguments to and from URL components.

    A ``Converter`` class should provide two instance methods:

    - ``to_string``: convert from a python object to a string
    - ``from_string``: convert from URL-encoded bytestring to the target
                        python type.

    It must also define the regular expression pattern that is used to extract
    the string from the URL.
    """

    pattern = "[^/]+"

    def __init__(self, pattern=None):
        """
        Initialize a ``Converter`` instance.
        """
        if pattern is not None:
            self.pattern = pattern

    def to_string(self, ob):
        """
        Convert arbitrary argument ``ob`` to a string representation
        """
        return str(ob)

    def from_string(self, s):
        """
        Convert string argument ``s`` to the target object representation,
        whatever that may be.
        """
        return s


class IntConverter(Converter):
    """\
    Match any integer value and convert to an ``int`` value.
    """

    pattern = r"[+-]?\d+"

    def from_string(self, s):
        """
        Return ``s`` converted to an ``int`` value.
        """
        return int(s)


class StrConverter(Converter):
    """\
    Match any string, not including a forward slash, and return a ``str``
    value
    """

    pattern = r"[^/]+"

    def to_string(self, ob):
        """
        Return ``s`` converted to an ``str`` object.
        """
        return ob

    def from_string(self, s):
        """
        Return ``s`` converted to a (unicode) string type
        """
        return s


class AnyConverter(StrConverter):
    """
    Match any one of the given string options. Example::

        "/<lang:any('fr', 'en', 'de')>"
    """

    def __init__(self, *args):
        super(AnyConverter, self).__init__(None)
        if len(args) == 0:
            raise ValueError("Must supply at least one argument to any()")
        self.pattern = "|".join(re.escape(arg) for arg in args)


class PathConverter(StrConverter):
    """\
    Match any string, possibly including forward slashes, and return a
    ``str`` object.
    """

    pattern = r".+"


class MatchAllURLsPattern(Pattern):
    """
    A pattern matcher that matches all paths starting with the given prefix.
    No arguments are parsed from the URL.
    """

    def __init__(self, path):
        self.path = path

    def match(self, path: str) -> t.Optional[PathMatch]:
        if path.startswith(self.path):
            return PathMatch(self.path, path[len(self.path) :], (), {})
        return None

    def pathfor(self, *args, **kwargs):
        assert (
            not args and not kwargs
        ), "MatchAllURLsPattern does not support URL arguments"

        return self.path, (), {}

    def path_argument_info(self):
        return (), {}

    def add_prefix(self, prefix):
        return self.__class__(join_path(prefix, self.path))

    def __str__(self):
        return "%s*" % (self.path,)


class PrefixPattern(MatchAllURLsPattern):
    """
    A pattern matcher that matches the given prefix. The prefix will only be
    matched at a path separator boundary.

    ``PrefixPattern('/foo')`` will match the paths ``/foo`` and ``/foo/bar``,
    but not ``/foobar``.

    No arguments are parsed from the URL.
    """

    def __init__(self, path):
        super().__init__(path)
        if path[-1] == "/":
            self.path_with_sep = path
        else:
            self.path_with_sep = f"{self.path}/"

    def match(self, path: str) -> t.Optional[PathMatch]:
        prefix = path[: len(self.path_with_sep)]

        if prefix == self.path or prefix == self.path_with_sep:
            return PathMatch(self.path, path[len(self.path) :], (), {})
        return None


class ExtensiblePattern(Pattern):
    """\
    An extensible URL pattern matcher.

    Synopsis::

        >>> from pprint import pprint
        >>> p = ExtensiblePattern(r"/<:str>/<year:int>/<title:str>")
        >>> pprint(p.match('/archive/1999/blah')) # doctest: +ELLIPSIS
        PathMatch(...)

    Patterns are split on slashes into components. A component can either be a
    literal part of the path, or a pattern component in the form::

        <identifier>:<converter>

    ``identifer`` can be any python name, which will be used as the name of a
    keyword argument to the matched function. If omitted, the argument will be
    passed as a positional arg.

    ``converter`` must be the name of a pre-registered converter. Converters
    must support ``to_string`` and ``from_string`` methods and are used to
    convert between URL segments and python objects.

    By default, the following converters are configured:

    - ``int`` - converts to an integer
    - ``path`` - any path (ie can include forward slashes)
    - ``str`` - any string (not including forward slashes)
    - ``unicode`` - alias for ``str``
    - ``any`` - a string matching a list of alternatives

    Some examples::

        >>> p = ExtensiblePattern(r"/images/<:path>")
        >>> p.match('/images/thumbnails/02.jpg') # doctest: +ELLIPSIS
        PathMatch(..., args=('thumbnails/02.jpg',), kwargs={})

        >>> p = ExtensiblePattern("/<page:any('about', 'help')>.html")
        >>> p.match('/about.html') # doctest: +ELLIPSIS
        PathMatch(..., args=(), kwargs={'page': 'about'})

        >>> p = ExtensiblePattern("/entries/<id:int>")
        >>> p.match('/entries/23')  # doctest: +ELLIPSIS
        PathMatch(..., args=(), kwargs={'id': 23})

    Others can be added by calling ``ExtensiblePattern.register_converter``
    """

    preset_patterns = {
        "int": IntConverter,
        "str": StrConverter,
        "unicode": StrConverter,
        "path": PathConverter,
        "any": AnyConverter,
    }
    pattern_parser = re.compile(
        r"""
        <
            (?P<name>\w[\w\d]*)?
            :
            (?P<converter>\w[\w\d]*)
            (?:
                \(
                         (?P<args>.*?)
                \)
            )?
        >
    """,
        re.X,
    )

    def __init__(self, pattern, match_entire_path=True):
        """
        Initialize a new ``ExtensiblePattern`` object with pattern ``pattern``

        :param pattern: The pattern string, eg ``'/<id:int>/show'``
        :param match_entire_path: Boolean. If ``True``, the entire path will be
                                  required to match, otherwise a prefix match
                                  will suffice.
        """
        super(ExtensiblePattern, self).__init__()

        self.pattern = pattern
        self.match_entire_path = match_entire_path

        self.segments = list(self._make_segments())
        self.args = [item for item in self.segments if item.converter is not None]
        self.segments_from_string = [
            (s.name, s.converter.from_string)
            for s in self.segments
            if s.converter is not None
        ]
        self.positional_args = tuple(a.converter for a in self.args if a.name is None)
        self.keyword_args = {
            a.name: a.converter for a in self.args if a.name is not None
        }
        regex = "".join(segment.regex for segment in self.segments)
        if self.match_entire_path:
            regex += "$"
        else:
            regex += "(?=/|$)"

        self.regex = re.compile(regex)
        self.regex_match = self.regex.match

    def path_argument_info(self):
        return (self.positional_args, self.keyword_args)

    def _make_segments(self):
        r"""
        Generate successive PatternSegment objects from the given string.

        Each segment object represents a part of the pattern to be matched, and
        comprises ``source``, ``regex``, ``name`` (if a named parameter) and
        ``converter`` (if a parameter)
        """
        for item in split_iter(self.pattern_parser, self.pattern):
            if isinstance(item, str):
                yield PatternSegment(item, re.escape(item), None, None)
                continue
            groups = item.groupdict()
            name, converter, args = (
                groups["name"],
                groups["converter"],
                groups["args"],
            )
            converter = self.preset_patterns[converter]
            if args:
                args, kwargs = self.parseargs(args)
                converter = converter(*args, **kwargs)
            else:
                converter = converter()
            yield PatternSegment(
                item.group(0), "(%s)" % converter.pattern, name, converter
            )

    def parseargs(self, argstr):
        """
        Return a tuple of ``(args, kwargs)`` parsed out of a string in the
        format ``arg1, arg2, param=arg3``.

        Synopsis::

            >>> ep =  ExtensiblePattern('')
            >>> ep.parseargs("1, 2, 'buckle my shoe'")
            ((1, 2, 'buckle my shoe'), {})
            >>> ep.parseargs("3, four='knock on the door'")
            ((3,), {'four': 'knock on the door'})

        """
        return eval("(lambda *args, **kwargs: (args, kwargs))(%s)" % argstr)

    def match(self, path) -> t.Optional[PathMatch]:
        """
        Test ``path`` and return a PathMatch object or ``None`` if there was no
        match.
        """
        mo = self.regex_match(path)
        if mo is None:
            return None
        groups = mo.groups()
        assert len(groups) == len(self.args), (
            "Number of regex groups does not match expected count. "
            "Perhaps you have used capturing parentheses somewhere? "
            "The pattern matched was %r." % self.regex.pattern
        )

        try:
            group_items = [
                (name, from_string(value))
                for value, (name, from_string) in zip(groups, self.segments_from_string)
            ]
        except ValueError:
            return None

        matched = mo.group(0)
        args = tuple(value for name, value in group_items if not name)
        kwargs = {name: value for name, value in group_items if name}
        return PathMatch(matched, path[len(matched) :], args, kwargs)

    def pathfor(
        self, *args, _strjoin="".join, **kwargs
    ) -> tuple[str, list[Any], dict[Any, Any]]:
        """
        Example usage::

            >>> p = ExtensiblePattern("/view/<name:str>/<revision:int>")
            >>> p.pathfor(name='important_document.pdf', revision=299)
            ('/view/important_document.pdf/299', [], {})

            >>> p = ExtensiblePattern("/view/<:str>/<:int>")
            >>> p.pathfor('important_document.pdf', 299)
            ('/view/important_document.pdf/299', [], {})
        """

        arg_list = list(args)
        result: list[str] = []
        result_append = result.append
        for seg in self.segments:
            if not seg.converter:
                result_append(seg.source)

            elif seg.name:
                try:
                    value = kwargs.pop(seg.name)
                except KeyError:
                    raise URLGenerationError(
                        "Argument %r not specified for url %r"
                        % (seg.name, self.pattern)
                    )
                result_append(seg.converter.to_string(value))

            else:
                try:
                    value = arg_list.pop(0)
                except IndexError:
                    raise URLGenerationError(
                        f"Not enough positional arguments for url {self.pattern}"
                    )
                result_append(seg.converter.to_string(value))

        return _strjoin(result), arg_list, kwargs

    def add_prefix(self, prefix):
        return self.__class__(join_path(prefix, self.pattern), self.match_entire_path)

    @classmethod
    def register_converter(cls, name, converter):
        r"""
        Register a preset pattern for later use in URL patterns.

        Example usage::

            >>> from datetime import date
            >>> from time import strptime
            >>> class DateConverter(Converter):
            ...     pattern = r'\d{8}'
            ...     def from_string(self, s):
            ...         return date(*strptime(s, '%d%m%Y')[:3])
            ...
            >>> ExtensiblePattern.register_converter('date', DateConverter)
            >>> ExtensiblePattern('/<:date>')\
            ...      .match('/01011970')  # doctest:+ELLIPSIS
            PathMatch(..., args=(datetime.date(1970, 1, 1),), kwargs={})
        """
        cls.preset_patterns[name] = converter

    def __repr__(self):
        return "<%s %r>" % (self.__class__, self.pattern)

    def __str__(self):
        return "%s" % (self.pattern,)


class PatternSegment:
    """
    Represent a single segment of a URL pattern, storing information about the
    ``source``, ``regex`` used to pattern match the segment, ``name`` for
    named parameters and the ``converter`` used to map the value to a URL
    parameter if applicable
    """

    __slots__ = ["source", "regex", "name", "converter"]

    def __init__(
        self,
        source: str,
        regex: str,
        name: Optional[str],
        converter: Optional[Converter],
    ):
        self.source = source
        self.regex = regex
        self.name = name
        self.converter = converter


class Route(object):
    """\
    Represent a URL routing pattern
    """

    #: The default class to use for URL pattern matching
    pattern_class = ExtensiblePattern

    fallthrough_statuses: Optional[set[int]]

    _route_hints: dict[Callable, dict[str, list[Callable]]] = defaultdict(
        lambda: defaultdict(list)
    )

    provide_request: Optional[bool] = None

    def __init__(
        self,
        pattern: t.Union[str, Pattern],
        methods: Optional[Collection[str]] = None,
        view: Optional[Union[ViewCallable, "RouteCollection", str]] = None,
        kwargs: Optional[dict[str, Any]] = None,
        args=None,
        name: Optional[str] = None,
        predicate=None,
        decorators=None,
        filters=None,
        fallthrough_on: Optional[Sequence[int]] = None,
        provide_request: Optional[bool] = None,
        **_kwargs: ViewCallable,
    ):
        """
        :param pattern:     A string that can be compiled into a path pattern
        :param methods:     The list of HTTP methods the view is bound to
                            ('GET', 'POST', etc)
        :param view:
            The view function, or a string identifier which will later be resolved.

        :param kwargs:      A dictionary of default keyword arguments to pass
                            to the view callable
        :param args:        Positional arguments to pass to the view callable
        :param name:        A name that can later be used to retrieve the route
                            for URL generation
        :param predicate:   A callable that is used to decide whether to match
                            this pattern. The callable must take a ``Request``
                            object as its only parameter and return a boolean.
        :param decorators:  Decorator functions to apply to the view callable
                            before invoking it
        :param filters:     Filter functions to apply to the view's return
                            value before returning the final response object
        :param fallthrough_on: A list of http status codes which, if returned
                               by a view will cause the current response to be
                               discarded with routing continuing to the next
                               available route.
        :param provide_request:
            If True, provide the current request as the first argument to the
            view callable. Defaults to ``None``, which will autodetect if the
            view function has an initial parameter of type
            :class:`~fresco.request.Request`
        :param **_kwargs:   Keyword arguments matching HTTP method names
                            (GET, POST etc) can used to specify views
                            associated with those methods.
                            Other keyword aruments are passed through to the
                            view callable.

        Naming routes
        -------------

        Naming routes allows you to reference routes throughout your
        application, eg when generating URLs with
        :function:`~fresco.core.urlfor`.

        If you don't specify a ``name`` argument, the route will available by
        its fully qualified function name (eg
        ``'package.module.view_function'``), or by passing the function object
        itself to ``urlfor``

        Views may often be assigned to multiple routes, for example::

            >>> def display_image(size):
            ...     return Response(['image data...'])
            ...
            >>> from fresco import FrescoApp
            >>> app = FrescoApp()
            >>> app.route('/image', display_image, size=(1024, 768))
            >>> app.route('/thumbnail', display_image, size=(75, 75))

        If you generate a URL for the view function ``display_image``, the
        the first declared route will always win, in this case::

            >>> app.urlfor(display_image)
            'http://localhost/image'

        To generate URLs for the thumbnail route in this example, you must
        explicity assign names::

            >>> app.route('/image', display_image, name='image')
            >>> app.route('/thumbnail', display_image, name='image-thumbnail')

            >>> app.urlfor('image')
            'http://localhost/image'
            >>> app.urlfor('image-thumbnail')
            'http://localhost/thumbnail'

        """
        method_view_map: dict[str, t.Union[ViewCallable, RouteCollection, str]] = {}
        if methods:
            assert view is not None
            if isinstance(methods, str):
                methods = [methods]
            else:
                # Catch the common error of not providing a valid method
                if not isinstance(methods, t.Iterable):
                    raise TypeError(
                        "HTTP methods must be specified as a string or iterable"
                    )
            for m in methods:
                if m == ALL:
                    methods = ALL_METHODS
                    break
                if m not in ALL_METHODS:
                    raise ValueError("{!r} is not a valid HTTP method".format(m))
            method_view_map.update((m, view) for m in methods)

        if view := _kwargs.pop("ALL", None):
            method_view_map |= {m: view for m in ALL_METHODS}

        for method in ALL_METHODS:
            if method in _kwargs:
                method_view_map[method] = _kwargs.pop(method)

        if not isinstance(pattern, Pattern):
            pattern = self.pattern_class(pattern)

        if name and ":" in name:
            # Colons are reserved to act as separators
            raise ValueError("Route names cannot contain ':'")

        self.name = name
        self.predicate = predicate
        self.decorators = decorators or []
        self.before_hooks: list[Callable] = []
        self.filters = filters or []
        if fallthrough_on:
            self.fallthrough_statuses = {int(i) for i in fallthrough_on}
        else:
            self.fallthrough_statuses = None

        if provide_request is not None:
            self.provide_request = provide_request

        #: Default values to use for path generation
        self.routed_args_default: dict[str, Any] = {}

        self.pattern = pattern
        self.methods = set(method_view_map)
        self.instance = None

        # Cached references to view functions
        self._cached_views: dict[str, ViewCallable] = {}

        # Cached references to decorated view function. We use weakrefs in case
        # a process_view hook substitutes the view function used as a key
        # for a dynamically generated function.
        self._cached_dviews: Mapping[Callable, Callable] = WeakKeyDictionary()

        p_args, p_kwargs = pattern.path_argument_info()
        for k in p_kwargs:
            default = _kwargs.pop(k + "_default", _marker)
            if default is not _marker:
                self.routed_args_default[k] = default

        self.view_args: tuple[Any] = tuple(
            args or _kwargs.pop("view_args", tuple())  # type: ignore
        )
        self.view_kwargs: dict[str, Any] = dict(
            kwargs or _kwargs.pop("view_kwargs", {}), **_kwargs  # type: ignore
        )

        for arg in self.view_args:
            if isinstance(arg, RouteArg):
                arg.configure(self, None)

        for argname, arg in self.view_kwargs.items():
            # Allow 'Route(... x=FormArg)' shortcut
            if isinstance(arg, type) and issubclass(arg, RouteArg):
                arg = self.view_kwargs[argname] = arg()
            if isinstance(arg, RouteArg):
                arg.configure(self, argname)

        #: A mapping of HTTP methods to view specifiers
        self.viewspecs = method_view_map

    def __repr__(self):
        view_methods_map: Mapping[
            t.Union[RouteCollection, ViewCallable, str], set[str]
        ] = defaultdict(set)
        for method, viewspec in self.viewspecs.items():
            view_methods_map[viewspec].add(method)

        s = []
        for viewspec, methods in view_methods_map.items():
            if methods == ALL_METHODS:
                method_str = "*"
            else:
                method_str = " ".join(self.methods)
            s.append("%s %s => %r" % (method_str, str(self.pattern), fq_path(viewspec)))

        return "<%s %s>" % (self.__class__.__name__, "\n      ".join(s))

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_cached_views"] = {}
        state["_cached_dviews"] = WeakKeyDictionary()
        return state

    def fallthrough_on(self, status_codes):
        newroute = copy(self)
        newroute.fallthrough_statuses = {int(s) for s in status_codes}
        return newroute

    def match(self, path: str, method: t.Optional[str]) -> t.Optional[PathMatch]:
        if method and method not in self.methods:
            return None
        return self.pattern.match(path)

    def getview(self, method: str) -> ViewCallable:
        """
        Resolve and return the raw view callable.
        """
        try:
            return self._cached_views[method]
        except KeyError:
            pass

        uview = self.viewspecs[method]
        if isinstance(uview, str):
            if "." in uview:
                mod_name, attr_name = uview.rsplit(".", 1)
                mod = import_module(mod_name)
                uview = getattr(mod, attr_name)

            else:
                if self.instance is None:
                    raise RouteNotReady()
                uview = getattr(self.instance, uview)

        self._cached_views[method] = uview  # type: ignore
        if self.provide_request is None:
            if callable(uview):
                self.provide_request = _has_request_parameter(uview)
            else:
                self.provide_request = False

        return uview  # type: ignore

    @classmethod
    def _add_route_hint(cls, viewfunc, hinttype, func):
        assert hinttype in {"before_hooks", "decorators", "filters"}
        cls._route_hints[viewfunc][hinttype].append(func)

    @classmethod
    def _get_route_hints(cls, func, hinttype):
        # Look up the underlying function in the case that view is a method
        viewfunc = getattr(func, "__func__", func)
        return cls._route_hints[viewfunc][hinttype]

    def getdecoratedview(self, view: Callable) -> Callable:
        """\
        Return the view callable decorated with any decorators defined in the
        route
        """
        try:
            return self._cached_dviews[view]
        except KeyError:
            pass

        hints = self._get_route_hints
        dview = view

        # Reverse order of before hooks: the last added should be run first.
        # This makes sense if extending existing routes with a before hook,
        # as generally you want your new hook to take precendence over any
        # existing hooks.
        before_hooks = hints(view, "before_hooks")[::-1] + self.before_hooks
        filters = self.filters + hints(view, "filters")
        decorators = self.decorators + hints(view, "decorators")

        for d in decorators:
            dview = d(dview)
        dview = self._add_before_hooks(dview, before_hooks)
        dview = self._add_filters(dview, filters)

        self._cached_dviews[view] = dview  # type: ignore
        return dview

    def _add_before_hooks(self, view, hooks):
        """
        Decorate ``view`` so that any 'before' hook functions are called.
        """
        if not hooks:
            return view

        def view_with_before_hooks(*args, **kwargs):
            for f in hooks:
                r = f(*args, **kwargs)
                if r is not None:
                    return r
            return view(*args, **kwargs)

        return view_with_before_hooks

    def _add_filters(self, view, filters):
        """
        Decorate ``view`` so that the result is passed through any available
        filters.
        """
        if not filters:
            return view

        def filtered_view(*args, **kwargs):
            result = view(*args, **kwargs)
            for f in filters:
                result = f(result)
            return result

        return filtered_view

    def bindto(self, instance):
        """\
        Return copy of the route bound to a given instance.

        Use this when traversing view classes.
        """
        ob = copy(self)
        ob._setinstance(instance)
        return ob

    def _setinstance(self, instance):
        self.instance = instance

    def add_prefix(self, path):
        """
        Return a copy of the Route object with the given path prepended to the
        routing pattern.
        """
        newroute = object.__new__(self.__class__)
        newroute.__dict__ = dict(self.__dict__, pattern=self.pattern.add_prefix(path))
        return newroute

    def path(self, *args, **kwargs):
        """\
        Build the path corresponding to this route and return a tuple of:

            ``(<path>, <remaining args>, <remaining kwargs>)``

        ``remaining args`` and ``remaining kwargs`` are any values from
        ``*args`` and ``**kwargs`` not consumed during path construction.

        See also :meth:`~fresco.routing.Pattern.pathfor`.
        """
        request = kwargs.pop("request", None)
        for k in self.routed_args_default:
            if k not in kwargs:
                v = self.routed_args_default[k]
                if callable(v):
                    v = v(request)
                kwargs[k] = v
        return self.pattern.pathfor(*args, **kwargs)

    def route_keys(self):
        """\
        Generate keys by which the route should be indexed
        """
        if self.name:
            yield self.name

        for method, viewspec in self.viewspecs.items():
            yield viewspec
            try:
                view = self.getview(method)
            except RouteNotReady:
                continue
            yield view

            # Also return the underlying function object for python2 unbound
            # methods
            func_ob = getattr(view, "__func__", None)
            if func_ob is not None:
                yield func_ob

    @classorinstancemethod
    def before(self, func, *args, **kwargs):
        """
        Call a function before passing the request to the view.

            Route('/view-secret-page', GET, view_secret_page)\
                    .before(check_logged_in)

        :param func: The function. Must accept the same arguments as the view
                     and may return either ``None`` or a
                     :class:`~fresco.response.Response` object.
                     If a response is returned the view will not be
                     invoked.

        :param args: Extra positional args to pass to ``func``
        :param kwargs: Extra keyword args to pass to ``func``
        """
        if args or kwargs:
            func = partial(func, *args, **kwargs)

        if isinstance(self, Route):
            self.before_hooks.append(func)
            return self

        else:

            def _decorator(view):
                self._add_route_hint(view, "before_hooks", func)
                return view

            return _decorator

    @classorinstancemethod
    def wrap(self, decorator, *args, **kwargs):
        """
        Wrap the view function in a decorator. Can be chained for a fluid api::

            Route('/user/<id:int>', GET, edituser)\
                    .wrap(require_ssl)
                    .wrap(logged_in)
        """
        if args or kwargs:
            decorator = partial(decorator, *args, **kwargs)

        if isinstance(self, Route):
            self.decorators.append(decorator)
            return self
        else:

            def _decorator(func):
                self._add_route_hint(func, "decorators", decorator)
                return func

            return _decorator

    #: ``decorate`` is an alias for :meth:`~fresco.routing.Route.wrap`
    decorate = wrap

    @classorinstancemethod
    def filter(self, func, *args, **kwargs):
        """
        Filter the output of the view function through other functions::

            Route('/user/<id:int>', GET, edituser)\
                    .wrap(require_ssl)
                    .filter(render, 'user.tmpl')

        :param func: The filter function. Must accept the output of the view
                     and return the filtered response
        :param args: Extra positional args to pass to ``func``
        :param kwargs: Extra keyword args to pass to ``func``
        """
        if args or kwargs:
            func = partial(func, *args, **kwargs)

        if isinstance(self, Route):
            self.filters.append(func)
            return self
        else:

            def _decorator(view):
                self._add_route_hint(view, "filters", func)
                return view

            return _decorator


class RRoute(Route):
    """
    A subclass of :class:`fresco.routing.Route` that always provides an initial
    `request` argument to the view.
    """

    provide_request = True


def split_iter(pattern, string):
    """
    Generate alternate strings and match objects for all occurences of
    ``pattern`` in ``string``.
    """
    matcher = pattern.finditer(string)
    match = None
    pos = 0
    for match in matcher:
        yield string[pos : match.start()]
        yield match
        pos = match.end()
    yield string[pos:]


def routefor(viewspec, _app=None):
    """\
    Convenience wrapper around :meth:`~fresco.routing.RouteCollection.urlfor`.
    """
    return context.app.routefor(viewspec)


class RouteCollection(MutableSequence):
    """\
    A collection of :class:`~fresco.routing.Route` objects, RouteCollections:

    - Contain methods to configure routes, including the ability to
      delegate URLs to other RouteCollections
    - Can map from a request to a view
    """

    route_class = Route

    #: The minimum size for the ``(method, path) => route`` cache
    min_route_cache_size = 20

    #: The maximum size for the ``(method, path) => route`` cache
    max_route_cache_size = 500

    _route_cache = None

    def __init__(self, routes=None, route_class=None, cache=True):
        self.__routes__: list[Route] = []
        self.__routed_views__: dict[Any, Union[Route, RouteNotFound]] = {}
        if cache:
            self.reinit_route_cache()
        self.route_class = route_class or self.route_class
        if routes is not None:
            for item in routes:
                if isinstance(item, Route):
                    self.add_route(item)
                elif hasattr(item, "__routes__"):
                    for r in item.__routes__:
                        if r.instance is None:
                            r = r.bindto(item)
                        self.add_route(r)
                elif isinstance(item, t.Iterable):
                    for r in item:
                        self.add_route(r)
                else:
                    raise TypeError(item)
        if cache:
            self.reinit_route_cache()

    def __add__(self, other):
        result = copy(self)
        if isinstance(other, Route):
            result.add_route(other)
        else:
            for item in other:
                result.add_route(item)
        return result

    def __radd__(self, other):
        if isinstance(other, Route):
            return RouteCollection([other]) + self
        elif isinstance(other, t.Iterable):
            return RouteCollection(other) + self
        raise TypeError("Cannot add %r to %r" % (other, self))

    def __iter__(self):
        return iter(self.__routes__)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.__class__(self.__routes__[index])
        return self.__routes__[index]

    def __setitem__(self, index, new):
        self.__routes__.__setitem__(index, new)
        if self._route_cache is not None:
            self.reinit_route_cache()

    def __delitem__(self, index):
        self.__routes__.__delitem__(index)
        if self._route_cache is not None:
            self.reinit_route_cache()

    def __len__(self):
        return len(self.__routes__)

    def reinit_route_cache(self):
        cache_size = max(
            min(self.max_route_cache_size, len(self.__routes__) * 4),
            self.min_route_cache_size,
        )
        self._route_cache = make_cache(self._get_routes, cache_size)

    def insert(self, index, value):
        self.__routes__.insert(index, value)
        if self._route_cache is not None:
            self.reinit_route_cache()

    def add_route(self, route):
        self.__routes__.append(route)
        if self._route_cache is not None:
            self.reinit_route_cache()

    def add_prefix(self, prefix):
        """
        Return a copy of the RouteCollection with the given path prepended to
        all routes.
        """
        ob = copy(self)
        ob.__routes__ = [route.add_prefix(prefix) for route in self]
        return ob

    def fallthrough_on(self, *status_codes):
        ob = copy(self)
        ob.__routes__ = [route.fallthrough_on(*status_codes) for route in self]
        return ob

    def pathfor(self, viewspec, *args, **kwargs):
        """\
        Return the path component of the url for the given view name/function
        spec.

        :param viewspec: a view name, a reference in the form
                         ``'package.module.viewfunction'``, or the view
                         callable itself.
        """
        request = kwargs.pop("request", None)
        if isinstance(viewspec, str) and ":" in viewspec:
            viewspec, remainder = viewspec.split(":", 1)
            delegated_route = self.routefor(viewspec)

            p1, remainder_args, remainder_kwargs = delegated_route.path(*args, **kwargs)

            factory_args = args[: -len(remainder_args)]
            factory_kwargs = {
                k: v for k, v in kwargs.items() if k not in remainder_kwargs
            }
            for k in delegated_route.routed_args_default:
                if k not in factory_kwargs:
                    v = delegated_route.routed_args_default[k]
                    if callable(v):
                        v = v(request)
                    factory_kwargs[k] = v
            rc = delegated_route.routecollectionfactory(  # type: ignore
                *factory_args, **factory_kwargs
            )
            p2 = rc.pathfor(remainder, request=request, *args, **kwargs)
            return p1 + p2

        return self.routefor(viewspec).path(request=request, *args, **kwargs)[0]

    def routefor(self, viewspec: Union[Callable, str]) -> Route:
        """
        Return the :class:`~fresco.routing.Route` instance associated with
        ``viewspec``.

        Views may have multiple routes bound, in this case the first bound
        route will always take precedence.

        This method does not resolve delegated routes.

        :param viewspec: a view callable or a string in the form
                         'package.module.viewfunction'
        """
        try:
            route_or_rnf = self.__routed_views__[viewspec]
            if isinstance(route_or_rnf, RouteNotFound):
                raise route_or_rnf
            return route_or_rnf
        except KeyError:
            pass

        for route in self.__routes__:
            for k in route.route_keys():
                self.__routed_views__.setdefault(k, route)

        try:
            route_or_rnf = self.__routed_views__[viewspec]
            if isinstance(route_or_rnf, RouteNotFound):
                raise route_or_rnf
            return route_or_rnf
        except KeyError:
            pass

        if not isinstance(viewspec, str):
            exc = self.__routed_views__[viewspec] = RouteNotFound(viewspec)
            raise exc

        modname, symbols = viewspec, []
        while True:
            try:
                modname, sym = modname.rsplit(".", 1)
            except ValueError:
                exc = self.__routed_views__[viewspec] = RouteNotFound(viewspec)
                raise exc

            symbols.append(sym)
            module = sys.modules.get(modname, None)
            if module:
                ob = module
                for s in reversed(symbols):
                    ob = getattr(ob, s)
                if callable(ob):
                    route_or_rnf = self.__routed_views__[ob]
                    self.__routed_views__[viewspec] = route_or_rnf
                    if isinstance(route_or_rnf, RouteNotFound):
                        raise route_or_rnf
                    return route_or_rnf
                else:
                    exc = self.__routed_views__[viewspec] = RouteNotFound(viewspec)
                    raise exc

    def _get_routes(
        self, key: tuple[t.Optional[str], str]
    ) -> t.Sequence[tuple[Route, PathMatch]]:
        method, path = key
        routes = ((r, r.match(path, method)) for r in self.__routes__)
        return [(r, t) for (r, t) in routes if t is not None]

    def get_route_traversals(
        self,
        path: str,
        method: Optional[str],
        request: Optional[Request] = None,
    ) -> t.Iterator[RouteTraversal]:
        """
        Generate RouteTraversals for routes matching the given path and
        method::

            for rt in routecollection.get_route_traversals('/foo/bar', GET):
                print("Route is", rt.route)
                print("Arguments extracted from the path:", rt.args, rt.kwargs)
                print("RouteCollections are:", rt.collections_traversed)

        :param path: the URL path to match (usually this is ``PATH_INFO``)
        :param method: the HTTP method to match (usually this is
                       ``REQUEST_METHOD``). May be ``None``, in which case
                       matching will be performed on the ``path`` argument
                       only.
        :param request: a :class:`~fresco.request.Request` object

        :return: A generator yielding ``RouteTraversal`` objects
        """
        if self._route_cache:
            routes, exc_info = self._route_cache((method, path))
            if exc_info is not None:
                raise exc_info[1].with_traceback(exc_info[2])
        else:
            routes = self._get_routes((method, path))

        for route, result in routes:
            if request and route.predicate and not route.predicate(request):
                continue

            # View function arguments extracted while traversing the path
            traversal_args, traversal_kwargs = result.args, result.kwargs
            if method is None:
                view = None
                for m in route.viewspecs:
                    view = route.getview(m)
                    break
            else:
                view = route.getview(method)

            # Process any args/kwargs defined in the Route declaration.
            if request:
                route_args = tuple(
                    (item(request) if isinstance(item, RouteArg) else item)
                    for item in route.view_args
                )

                route_kwargs = {
                    k: v(request) if isinstance(v, RouteArg) else v
                    for k, v in route.view_kwargs.items()
                }
            else:
                route_args = route.view_args
                route_kwargs = route.view_kwargs

            args = route_args + traversal_args
            if route.provide_request:
                args = (request,) + args
            kwargs = dict(traversal_kwargs, **route_kwargs)

            if isinstance(route, DelegateRoute):
                try:
                    sub_routes = route.routecollectionfactory(*args, **kwargs)
                except ResponseException as e:
                    exc = e

                    def raiser(*args, **kwargs):
                        raise exc

                    r = self.route_class("/", ALL_METHODS, raiser)
                    yield RouteTraversal(r, r.getview("GET"), (), {}, [(self, "", r)])
                    continue

                for sub in sub_routes.get_route_traversals(
                    result.path_remaining, method, request
                ):
                    traversed = [
                        TraversedCollection(
                            self,
                            "",
                            route,
                            args,
                            kwargs,
                            traversal_args,
                            traversal_kwargs,
                        )
                    ]
                    traversed.extend(
                        TraversedCollection(c, result.path_matched + p, r, a, k, ta, tk)
                        for c, p, r, a, k, ta, tk in sub.collections_traversed
                    )

                    # Dynamic routes consume their arguments when creating the
                    # sub RouteCollection.
                    if route.dynamic:
                        yield RouteTraversal(
                            sub.route, sub.view, sub.args, sub.kwargs, traversed
                        )
                    else:
                        yield RouteTraversal(
                            sub.route,
                            sub.view,
                            args + sub.args,
                            kwargs | sub.kwargs,
                            traversed,
                        )
            else:
                yield RouteTraversal(
                    route,
                    view,
                    args,
                    kwargs,
                    [
                        TraversedCollection(
                            self,
                            "",
                            route,
                            args,
                            kwargs,
                            traversal_args,
                            traversal_kwargs,
                        )
                    ],
                )

    def get_routes(
        self, path: str, method: str, request: Optional[Request] = None
    ) -> t.Iterable[RouteTraversal]:
        warnings.warn(
            "RouteCollection.get_routes is deprecated and will be removed "
            "in a future version. Use get_route_traversals instead",
            DeprecationWarning,
        )
        return self.get_route_traversals(path, method, request)

    def route(
        self,
        pattern: t.Union[str, Pattern],
        methods: t.Optional[t.Union[str, t.Collection[str]]] = None,
        view: t.Optional[t.Union[str, ViewCallable]] = None,
        *args,
        route_class: t.Optional[t.Type[Route]] = None,
        **kwargs,
    ):
        """
        Match a URL pattern to a view function. Can be used as a function
        decorator, in which case the ``view`` parameter should not be passed.

        :param pattern: A string that can be compiled into a path pattern
        :param methods: A list of HTTP methods the view is bound to
        :param view:    The view function. If not specified a function
                        decorator will be returned.

        Other parameters are as for the :class:`Route` constructor.
        """
        # Catch the common error of not providing a valid method
        if methods and not isinstance(methods, t.Iterable):
            raise TypeError("HTTP methods must be specified as a string or iterable")

        # Called as a decorator?
        if methods is not None and view is None:

            def route_decorator(func):
                self.add_route(
                    self.route_class(pattern, methods, func, *args, **kwargs)
                )
                return func

            return route_decorator

        cls = route_class if route_class is not None else self.route_class
        route = cls(pattern, methods, view, *args, **kwargs)
        self.add_route(route)
        return route

    def route_wsgi(
        self,
        path: str,
        wsgiapp: t.Union[WSGIApplication, str],
        rewrite_script_name: bool = True,
        *args,
        **kwargs,
    ):
        """
        Route requests to ``path`` to the given WSGI application

        :param path: the mount point for the app
        :param wsgiapp: the WSGI callable, or the path to a callable in the
                        format ``myapp.module.do_wsgi``.
        :param rewrite_script_name: if ``True`` (the default),
                                    the mount point specified by ``path`` will
                                    be shifted off ``PATH_INFO`` and into the
                                    ``SCRIPT_NAME`` environ key.
        """

        def fake_start_response(status, headers, exc_info=None):
            return None

        if path.encode("ascii", "ignore").decode("ascii") == path:
            ws_path = path
        else:
            ws_path = None

        resolved_wsgi_app = None

        def fresco_wsgi_view(
            request: Request,
            path=path,
            ws_path=ws_path,
            rewrite_script_name=rewrite_script_name,
            fake_start_response=fake_start_response,
            make_response=Response.from_wsgi,
        ):
            nonlocal resolved_wsgi_app
            environ = request.environ
            ws_SCRIPT_NAME = environ["SCRIPT_NAME"]
            shift_script_name = rewrite_script_name and not (
                path == "/" and ws_SCRIPT_NAME == ""
            )
            if shift_script_name:
                if ws_path is None:
                    ws_path = path.encode(request.charset).decode("ISO-8859-1")
                environ = environ.copy()
                del environ[request.STATE_ENV_KEY]
                environ["SCRIPT_NAME"] = ws_SCRIPT_NAME + ws_path
                environ["PATH_INFO"] = environ["PATH_INFO"][len(ws_path) :]

            if resolved_wsgi_app is None:
                if isinstance(wsgiapp, str):
                    if "." in wsgiapp:
                        mod_name, attr_name = wsgiapp.rsplit(".", 1)
                        mod = import_module(mod_name)
                        resolved_wsgi_app = getattr(mod, attr_name)
                else:
                    resolved_wsgi_app = wsgiapp

            return make_response(resolved_wsgi_app, environ, fake_start_response)

        return self.route_all(
            path, ALL_METHODS, fresco_wsgi_view, route_class=RRoute, *args, **kwargs
        )

    def route_all(
        self, path, methods=None, view=None, *args, route_class=None, **kwargs
    ):
        """
        Expose a view for all URLs starting with ``path``.

        :param path: the path prefix at which the view will be routed
        """
        return self.route(
            PrefixPattern(path),
            methods,
            view,
            route_class=route_class,
            *args,
            **kwargs,
        )

    def include(self, path, views, fallthrough_on=None):
        """
        Include a view collection at the given path.

        The included view collection's url properties will be modified to
        generate the prefixed URLs.

        :param path:  Path at which to include the views
        :param views: Any collection of views (must expose a ``__routes___``
                      attribute)
        :param fallthrough_on: a fallthrough_on argument,
                                passed to :class:~`fresco.routing.Route`
        """
        routes = list(r.add_prefix(path) for r in views.__routes__)
        for r in routes:
            if r.instance is None:
                r = r.bindto(views)
            if fallthrough_on:
                r = r.fallthrough_on(fallthrough_on)
            self.add_route(r)

    def delegate(self, path, app, dynamic=False, *args, **kwargs):
        """\
        Delegate all requests under ``path`` to ``app``

        :param path: the path prefix at which the app will be available
        :param app: a FrescoApp instance
        """
        route = DelegateRoute(path, app, dynamic, *args, **kwargs)
        self.add_route(route)
        return route

    def replace(self, viewspec, newroute):
        """
        Replace the route(s) identified by ``viewspec`` with a new
        Route.

        :param viewspec: The routed view callable, or its fully qualified name
                         ('package.module.view_function'), or the name of a
                         named route
        :param newroute: The replacement route. This may be None, in which
                         case the route will be removed without replacement
        """
        route = self.routefor(viewspec)

        if newroute:
            position = self.__routes__.index(route)
            self.__routes__[position] = newroute
        else:
            self.__routes__.remove(route)

        for k, r in list(self.__routed_views__.items()):
            if r is route:
                del self.__routed_views__[k]

    def remove(self, value):
        """
        Remove the route(s) identified by ``viewspec``

        :param viewspec: The routed view callable, or its fully qualified name
                         ('package.module.view_function'), or the name of a
                         named route
        """
        self.replace(value, None)


class DelegateRoute(Route):
    """\
    A specialisation of Route that is used to delegate a path prefix to another
    route collection.
    """

    def __init__(self, prefix, view, dynamic=False, *args, **kwargs):
        pattern = ExtensiblePattern(prefix, match_entire_path=False)
        self.dynamic = dynamic
        if self.dynamic:
            self.routecollectionfactory = self._dynamic_routecollectionfactory
        else:
            self.routecollectionfactory = self._static_routecollectionfactory
            if not isinstance(view, RouteCollection):
                routes = [
                    r.bindto(view) if r.instance is None else r for r in view.__routes__
                ]
                view = RouteCollection(routes)
        super(DelegateRoute, self).__init__(pattern, ALL_METHODS, view, *args, **kwargs)

    def _dynamic_routecollectionfactory(self, *args, **kwargs) -> RouteCollection:
        """\
        Return the RouteCollection responsible for paths under this route
        """

        routes = self.getdecoratedview(self.getview(GET))(*args, **kwargs)

        if isinstance(routes, RouteCollection):
            return routes

        return RouteCollection(
            (r.bindto(routes) for r in routes.__routes__), cache=False
        )

    def _static_routecollectionfactory(self, *args, **kwargs) -> RouteCollection:
        return self.getview(GET)  # type: ignore


def register_converter(name, registry=ExtensiblePattern):
    """
    Class decorator that registers a converter class for use with
    ExtensiblePattern.

    Example::

        >>> @register_converter('hex')
        ... class HexStringConverter(Converter):
        ...     pattern = r'[a-f0-9]+'
        ...
    """

    def register_converter(cls):
        registry.register_converter(name, cls)
        return cls

    return register_converter


def _has_request_parameter(func: Callable[..., Any]) -> bool:
    """
    Return True if the given function has an initial parameter of type Request
    """

    def _is_request_annotation(a: Any) -> bool:
        if isinstance(a, type) and issubclass(a, Request):
            return True

        origin = t.get_origin(a)
        if origin is t.Union and any(
            _is_request_annotation(arg) for arg in t.get_args(a)
        ):
            return True
        return False

    sig = inspect.signature(func)
    firstparam = next(iter(sig.parameters.values()), None)
    return bool(firstparam and _is_request_annotation(firstparam.annotation))



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
Utilities for interfacing with WSGI
"""

import logging
from functools import partial
from io import BytesIO
from urllib.parse import unquote
from urllib.parse import urlparse
from typing import List
import typing as t
import sys

from fresco.types import WSGIApplication
from fresco.types import WriteCallable
from fresco.types import OptionalExcInfo
from fresco.types import HeaderList

logger = logging.getLogger(__name__)
T = t.TypeVar("T")


__all__ = [
    "environ_to_str",
    "str_to_environ",
    "environ_to_bytes",
    "bytes_to_environ",
    "StartResponseWrapper",
    "ClosingIterator",
    "make_environ",
]

#: The core keys expected in the WSGI environ dict, as defined by PEP3333
WSGI_KEYS = set(
    [
        "REQUEST_METHOD",
        "SCRIPT_NAME",
        "PATH_INFO",
        "QUERY_STRING",
        "CONTENT_TYPE",
        "CONTENT_LENGTH",
        "SERVER_NAME",
        "SERVER_PORT",
        "SERVER_PROTOCOL",
    ]
)


#: List of standard request header field names
#: Taken from the list available at
#: http://en.wikipedia.org/wiki/List_of_HTTP_header_fields which combines
#: RFC specified headers with commonly used non-standard header names
REQUEST_HEADER_NAMES = {
    "accept": "Accept",
    "accept_charset": "Accept-Charset",
    "accept_encoding": "Accept-Encoding",
    "accept_language": "Accept-Language",
    "accept_datetime": "Accept-Datetime",
    "authorization": "Authorization",
    "cache_control": "Cache-Control",
    "connection": "Connection",
    "cookie": "Cookie",
    "content_length": "Content-Length",
    "content_md5": "Content-MD5",
    "content_type": "Content-Type",
    "date": "Date",
    "expect": "Expect",
    "from": "From",
    "host": "Host",
    "if_match": "If-Match",
    "if_modified_since": "If-Modified-Since",
    "if_none_match": "If-None-Match",
    "if_range": "If-Range",
    "if_unmodified_since": "If-Unmodified-Since",
    "max_forwards": "Max-Forwards",
    "origin": "Origin",
    "pragma": "Pragma",
    "proxy_authorization": "Proxy-Authorization",
    "range": "Range",
    "referer": "Referer",
    "te": "TE",
    "user_agent": "User-Agent",
    "upgrade": "Upgrade",
    "via": "Via",
    "warning": "Warning",
    "common": "Common",
    "field": "Field",
    "x_requested_with": "X-Requested-With",
    "dnt": "DNT",
    "x_forwarded_for": "X-Forwarded-For",
    "x_forwarded_host": "X-Forwarded-Host",
    "x_forwarded_proto": "X-Forwarded-Proto",
    "front_end_https": "Front-End-Https",
    "x_http_method_override": "X-Http-Method-Override",
    "x_att_deviceid": "X-ATT-DeviceId",
    "x_wap_profile": "X-Wap-Profile",
    "proxy_connection": "Proxy-Connection",
}


def environ_to_str(s, enc="UTF-8"):
    """
    Decode a WSGI environ value to a string
    """
    return s.encode("iso-8859-1").decode(enc)


def str_to_environ(s, enc="UTF-8"):
    """
    Return a string encoded for a WSGI environ value

    This returns a 'bytes-as-unicode' string:

    - encode ``s`` using the specified encoding (eg utf8)
    - decode the resulting byte string as latin-1
    """
    return s.encode(enc, "surrogateescape").decode("iso-8859-1")


def unicode_to_environ(s, enc="UTF-8"):
    import warnings

    warnings.warn(
        "unicode_to_environ is deprecated, use environ_to_str instead",
        DeprecationWarning,
    )
    return environ_to_str(s, enc)


def environ_to_bytes(s):
    """
    Decode a WSGI environ value to a bytes object
    """
    return s.encode("latin1")


def bytes_to_environ(s):
    """
    Encode a byte string to a WSGI environ value

    This returns a 'bytes-as-unicode' string.
    """
    return s.decode("latin1")


def getenv(environ, key, default=None):
    """
    Return the named ``key`` from the WSGI environ dict ``environ`` as
    a byte string.
    """
    try:
        return environ[key].encode("ISO-8859-1")
    except KeyError:
        return default


def setenv(environ, key, value):
    """
    Set the named ``key`` to ``value`` in the WSGI environ dict ``environ``.
    Value must be a byte ``str`` and will be encoded according
    to the rules in PEP-3333.
    """
    environ[key] = value.decode("ISO-8859-1")


class StartResponseWrapper(object):
    """\
    Wrapper class for the ``start_response`` callable, allowing middleware
    applications to intercept and interrogate the proxied start_response
    arguments.

    Synopsis::

        >>> def my_wsgi_app(environ, start_response):
        ...     start_response('200 OK', [('Content-Type', 'text/plain')])
        ...     return ['Whoa nelly!']
        ...
        >>> def my_other_wsgi_app(environ, start_response):
        ...     responder = StartResponseWrapper(start_response)
        ...     result = my_wsgi_app(environ, responder)
        ...     print("Got status", responder.status)
        ...     print("Got headers", responder.headers)
        ...     responder.call_start_response()
        ...     return result
        ...
        >>> from flea import Agent
        >>> result = Agent(my_other_wsgi_app).get('/')
        Got status 200 OK
        Got headers [('Content-Type', 'text/plain')]

    See also ``Response.from_wsgi``, which takes a wsgi callable, environ and
    start_response, and returns a ``Response`` object, allowing the client to
    further interrogate and customize the WSGI response.

    Note that it is usually not advised to use this directly in middleware as
    start_response may not be called directly from the WSGI application, but
    rather from the iterator it returns. In this case the middleware may need
    logic to accommodate this. It is usually safer to use
    ``Response.from_wsgi``, which takes this into account.
    """

    def __init__(self, start_response):
        self.start_response = start_response
        self.status = None
        self.headers: List[str] = []
        self.called = False
        self.buf = BytesIO()
        self.exc_info = None

    def __call__(self, status, headers, exc_info=None):
        """
        Dummy WSGI ``start_response`` function that stores the arguments for
        later use.
        """
        self.status = status
        self.headers = headers
        self.exc_info = exc_info
        self.called = True
        return self.buf.write

    def call_start_response(self):
        """
        Invoke the wrapped WSGI ``start_response`` function.
        """
        try:
            write = self.start_response(self.status, self.headers, self.exc_info)
            write(self.buf.getvalue())
            return write
        finally:
            # Avoid dangling circular ref
            self.exc_info = None


class ClosingIterator(t.Generic[T]):
    """
    Wrap a WSGI iterator to allow additional close functions to be called on
    application exit.

    Synopsis::

        >>> class filelikeobject(object):
        ...
        ...     def read(self):
        ...         print("file read!")
        ...         return ''
        ...
        ...     def close(self):
        ...         print("file closed!")
        ...
        >>> def app(environ, start_response):
        ...     f = filelikeobject()
        ...     start_response('200 OK', [('Content-Type', 'text/plain')])
        ...     return ClosingIterator(iter(f.read, ''), f.close)
        ...
        >>> from flea import Agent
        >>> m = Agent(app).get('/')
        file read!
        file closed!

    """

    __slots__ = ("_iterable", "_next", "_close_funcs", "_closed")

    def __init__(self, iterable: t.Iterable[T], *close_funcs: t.Callable[[], None]):
        """
        Initialize a ``ClosingIterator`` to wrap iterable ``iterable``, and
        call any functions listed in ``*close_funcs`` on the instance's
        ``.close`` method.
        """
        self._iterable = iterable
        self._next = partial(next, iter(iterable))
        self._close_funcs = close_funcs
        iterable_close = getattr(self._iterable, "close", None)
        if iterable_close is not None:
            self._close_funcs = (iterable_close,) + close_funcs
        self._closed = False

    def __iter__(self):
        """
        ``__iter__`` method
        """
        return self

    def __next__(self):
        """
        Return the next item from the iterator
        """
        return self._next()

    def close(self):
        """
        Call all close functions listed in ``*close_funcs``.
        """
        for func in self._close_funcs:
            try:
                func()
            except Exception:
                logger.exception(
                    f"ContentIterator close function {func!r} raised an exception"
                )
        self._closed = True

    def __del__(self):
        """
        Emit a warning if the iterator is deleted without ``close`` having been
        called.
        """
        if not self._closed:
            try:
                import warnings
            except ImportError:
                return
            else:
                warnings.warn("%r deleted without close being called" % (self,))


def make_environ(url="/", environ=None, wsgi_input=b"", **kwargs):
    """
    Return a WSGI environ dict populated with values modelling the specified
    request url and data.

    :param url: The URL for the request,
                eg ``/index.html`` or ``/search?q=foo``.
                Note that ``url`` must be properly URL encoded.
    :param environ: values to pass into the WSGI environ dict
    :param wsgi_input: The input stream to use in the ``wsgi.input``
                        key of the environ dict
    :param kwargs: additional keyword arguments will be passed into the
                    WSGI request environment
    """
    url = urlparse(url)
    netloc = url.netloc
    user = ""
    if "@" in netloc:
        user, netloc = netloc.split("@", 1)

    if ":" in user:
        user = user.split(":")[0]

    if isinstance(wsgi_input, bytes):
        wsgi_input = BytesIO(wsgi_input)

    env_overrides = environ or {}
    for key, value in kwargs.items():
        key = key.replace("wsgi_", "wsgi.")
        if "." not in key:
            if value is None:
                continue
            # Convert core WSGI keys to upper case
            if key.upper() in WSGI_KEYS:
                key = key.upper()
            # Convert header names to form HTTP_USER_AGENT
            elif key.lower() in REQUEST_HEADER_NAMES:
                key = "HTTP_" + key.upper()
            # value must be a python native str, whatever that means
            if not isinstance(value, str):
                value = str_to_environ(value)
        env_overrides[key] = value

    environ = {
        "REQUEST_METHOD": "GET",
        "SERVER_NAME": "localhost",
        "SERVER_PORT": "80",
        "SERVER_PROTOCOL": "HTTP/1.0",
        "HTTP_HOST": netloc or "localhost",
        "SCRIPT_NAME": "",
        "PATH_INFO": str_to_environ(unquote(url.path)),
        "REMOTE_ADDR": "127.0.0.1",
        "wsgi.version": (1, 0),
        "wsgi.url_scheme": url.scheme or "http",
        "wsgi.input": wsgi_input,
        "wsgi.errors": sys.stderr,
        "wsgi.multithread": True,
        "wsgi.multiprocess": True,
        "wsgi.run_once": False,
    }
    if url.scheme == "https":
        environ["HTTPS"] = "on"
        environ["SERVER_PORT"] = "443"

    if user:
        environ["REMOTE_USER"] = user

    if url.query:
        environ["QUERY_STRING"] = url.query

    environ.update(env_overrides)
    for k, v in kwargs.items():
        if v is None and "." not in k:
            del environ[k]
    return environ


def apply_request(
    request, wsgicallable: WSGIApplication
) -> tuple[str, HeaderList, OptionalExcInfo, list[bytes]]:
    """
    Execute ``wsgicallable`` with the given request, exhaust and close the
    content iterator and return the result.
    """

    _start_response_result: list[tuple[str, HeaderList, OptionalExcInfo]] = []

    def start_response(
        status: str, headers: HeaderList, exc_info: OptionalExcInfo = None
    ) -> WriteCallable:
        _start_response_result.append((status, headers, exc_info))
        return lambda s: None

    contentiter = wsgicallable(request.environ, start_response)
    assert len(_start_response_result) == 1
    content = list(contentiter)
    if hasattr(contentiter, "close"):
        contentiter.close()  # type: ignore
    return _start_response_result[0] + (content,)



# encoding: utf-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from copy import copy
from functools import wraps
from functools import partial
from unittest.mock import Mock, call
import typing as t

import pytest
import tms

from fresco import FrescoApp
from fresco.core import urlfor
from fresco.request import Request
from fresco.exceptions import NotFound
from fresco.response import Response
from fresco.routing import ALL_METHODS
from fresco.routing import ALL
from fresco.routing import GET
from fresco.routing import OPTIONS
from fresco.routing import POST
from fresco.routing import _has_request_parameter
from fresco.routing import (
    Route,
    DelegateRoute,
    RouteCollection,
    routefor,
    RouteTraversal,
    RouteNotFound,
    TraversedCollection,
    register_converter,
    Converter,
)
from . import fixtures


def dummyview() -> Response:
    return Response()


def assert_method_bound_to(method, ob):
    try:
        assert method.__self__ is ob
    except AttributeError:
        assert method.im_self is ob


class TestMethodDispatch(object):
    def test_route_is_dispatched_to_correct_method(self):
        getview = Mock(return_value=Response())
        postview = Mock(return_value=Response())
        app = FrescoApp()
        app.route("/", GET, getview)
        app.route("/", POST, postview)

        with app.requestcontext("/"):
            app.view()
            assert getview.call_count == 1
            assert postview.call_count == 0

        with app.requestcontext_post("/"):
            app.view()
            assert getview.call_count == 1
            assert postview.call_count == 1


class TestRouteConstructor(object):
    def test_multiple_views_can_be_associated_with_a_route(self):
        app = FrescoApp()
        v1 = Mock(return_value=Response())
        v2 = Mock(return_value=Response())
        app.route("/", GET=v1, POST=v2)

        with app.requestcontext():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 0

        with app.requestcontext_post():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 1

    def test_kwargs_take_precedence(self):
        app = FrescoApp()
        v1 = Mock(return_value=Response())
        v2 = Mock(return_value=Response())
        app.route("/", ALL_METHODS, v1, POST=v2)

        with app.requestcontext():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 0

        with app.requestcontext_post():
            app.view()
            assert v1.call_count == 1
            assert v2.call_count == 1

    def test_it_catches_invalid_methods(self):
        # Route must be a string or an iterable
        with pytest.raises(TypeError):
            Route("/", object(), lambda: None)

        # Route must be a valid HTTP method
        with pytest.raises(ValueError):
            Route("/", "FOO", lambda: None)

        with pytest.raises(ValueError):
            Route("/", ["GET", "FOO"], lambda: None)

    def test_ALL_kwarg_routes_all_methods(self):
        def view():
            return Response()

        app = FrescoApp()
        app.route("/x", ALL, view)
        assert len(list(app.get_route_traversals("/x", GET))) == 1
        assert len(list(app.get_route_traversals("/x", POST))) == 1
        assert len(list(app.get_route_traversals("/x", OPTIONS))) == 1

        app.route("/y", ALL=view)
        assert len(list(app.get_route_traversals("/y", GET))) == 1
        assert len(list(app.get_route_traversals("/y", POST))) == 1
        assert len(list(app.get_route_traversals("/y", OPTIONS))) == 1


class TestRouteBeforeHooks(object):
    def test_hook_is_called(self):
        mock = Mock(return_value=None)
        app = FrescoApp([Route("/", GET=Response).before(lambda: mock())])

        with app.requestcontext("/"):
            app.view()
        assert mock.call_count == 1

    def test_hook_is_applied_with_decorator_syntax(self):
        mock = Mock(return_value=None)

        @Route.before(lambda: mock())
        def view():
            return Response("foo")

        app = FrescoApp([Route("/", GET=view)])
        with app.requestcontext("/"):
            app.view()
        assert mock.call_count == 1

    def test_it_calls_view_if_none_returned(self):
        viewmock = Mock(return_value=Response("view response"))
        hookmock = Mock(return_value=None)
        app = FrescoApp([Route("/", GET=viewmock).before(hookmock)])
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"view response"]
        assert viewmock.call_count == 1

    def test_it_aborts_view_if_value_returned(self):
        viewmock = Mock(return_value=Response("view response"))
        hookmock = Mock(return_value=Response("hook response"))
        app = FrescoApp([Route("/", GET=viewmock).before(hookmock)])
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"hook response"]

        assert viewmock.call_count == 0

    def test_it_passes_args_to_hook(self):
        mock = Mock(return_value=None)
        app = FrescoApp([Route("/", GET=Response, foo="bar").before(mock)])
        with app.requestcontext("/"):
            app.view()
        assert mock.call_args_list == [call(foo="bar")]

    def test_it_applies_chained_hook_calls_in_order(self):
        mock = Mock(return_value=None)

        @Route.before(mock, 1)
        @Route.before(mock, 2)
        def view():
            return Response()

        app = FrescoApp()
        app.route("/", GET, view).before(mock, 3).before(mock, 4)

        with app.requestcontext("/"):
            app.view()
        assert mock.call_args_list == [call(1), call(2), call(3), call(4)]


class TestRouteViewFilters(object):
    def exclaim(self, response):
        return response.replace(content=[b"".join(response.content_iterator) + b"!"])

    def ask(self, response):
        return response.replace(content=[b"".join(response.content_iterator) + b"?"])

    def test_filter_is_applied(self):
        views = fixtures.CBV("test")
        app = FrescoApp()
        app.route("/", GET, views.index_html).filter(self.ask)
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"test?"]

    def test_filter_is_applied_as_route_kwargs(self):
        views = fixtures.CBV("test")
        app = FrescoApp()
        app.route("/", GET, views.index_html, filters=[self.ask])
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"test?"]

    def test_it_passes_args_to_filter(self):
        app = FrescoApp()
        views = fixtures.CBV("test")

        def filter_func(r, s):
            return r.replace(content=r.content + [s])

        app.route("/", GET, views.index_html).filter(filter_func, s="foo")
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"test", b"foo"]

    def test_it_applies_chained_filter_calls_in_order(self):
        app = FrescoApp()
        views = fixtures.CBV("test")
        app.route("/", GET, views.index_html).filter(self.ask).filter(self.exclaim)
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"test?!"]

    def test_filter_is_applied_with_decorator_syntax(self):
        @Route.filter(lambda c: Response(c + "!"))
        def view():
            return "foo"

        app = FrescoApp()
        app.route("/", GET, view)
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"foo!"]


class TestRouteDecorators(object):
    def exclaim(self, func):
        @wraps(func)
        def exclaim(*args, **kwargs):
            response = func(*args, **kwargs)
            return response.replace(
                content=[b"".join(response.content_iterator) + b"!"]
            )

        return exclaim

    def test_decorator_is_applied(self):
        views = fixtures.CBV("test")

        app = FrescoApp()
        app.route("/decorated", GET, views.index_html, decorators=[self.exclaim])
        app.route("/plain", GET, views.index_html)

        with app.requestcontext("/decorated"):
            assert list(app.view().content_iterator) == [b"test!"]

        with app.requestcontext("/plain"):
            assert list(app.view().content_iterator) == [b"test"]

    def test_decorator_is_applied_with_wrap_method(self):
        views = fixtures.CBV("test")

        app = FrescoApp()
        app.route("/decorated", GET, views.index_html).wrap(self.exclaim)
        app.route("/plain", GET, views.index_html)

        with app.requestcontext("/decorated"):
            assert list(app.view().content_iterator) == [b"test!"]

        with app.requestcontext("/plain"):
            assert list(app.view().content_iterator) == [b"test"]

    def test_decorator_is_applied_with_decorator_syntax(self):
        @Route.decorate(self.exclaim)
        def f():
            return Response(["test"])

        class A(object):
            __routes__ = [Route("/", GET, "view")]

            @Route.decorate(self.exclaim)
            def view(self):
                return Response(["test"])

        app = FrescoApp()
        app.route("/", GET, f)
        app.include("/a", A())
        with app.requestcontext("/"):
            assert list(app.view().content_iterator) == [b"test!"]
        with app.requestcontext("/a/"):
            assert list(app.view().content_iterator) == [b"test!"]

    def test_decorator_works_with_urlfor(self):
        views = fixtures.CBV("test")
        app = FrescoApp()
        app.route("/decorated", GET, views.index_html, decorators=[self.exclaim])
        with app.requestcontext():
            assert urlfor(views.index_html, _app=app) == "http://localhost/decorated"

    def test_using_wraps_with_viewspec_doesnt_raise_AttributeError(self):
        def decorator(func):
            @wraps(func)
            def decorator(*args, **kwargs):
                return func(*args, **kwargs)

            return decorator

        class Views(object):
            __routes__ = (Route("/", GET, "index_html", decorators=[decorator]),)

            def index_html(self):
                return Response(["hello"])

        app = FrescoApp()
        app.include("/", Views())


class TestPredicates(object):
    def test_predicate_match(self):
        def v1():
            return Response([b"x"])

        def v2():
            return Response([b"y"])

        app = FrescoApp()
        app.route("/", GET, v1, predicate=lambda request: "x" in request.query)
        app.route("/", GET, v2, predicate=lambda request: "y" in request.query)

        with app.requestcontext("/?x=1"):
            assert b"".join(app.view().content) == b"x"
        with app.requestcontext("/?y=1"):
            assert b"".join(app.view().content) == b"y"
        with app.requestcontext("/"):
            assert app.view().status_code == 404


class TestRouteNames(object):
    def test_name_present_in_route_keys(self):
        r = Route("/", GET, lambda: None, name="foo")
        assert "foo" in list(r.route_keys())

    def test_name_with_other_kwargs(self):
        r = Route("/", GET, lambda: None, name="foo", x="bar")
        assert "foo" in list(r.route_keys())

    def test_name_cannot_contain_colon(self):
        with pytest.raises(ValueError):
            Route("/", GET, lambda: None, name="foo:bar")


class TestRouteCollection(object):
    def test_it_adds_routes_from_constructor(self):
        r1 = Route("/1", GET, dummyview, name="1")
        r2 = Route("/2", POST, dummyview, name="2")
        rc = RouteCollection([r1, r2])
        assert [r.name for r in rc] == ["1", "2"]

    def test_it_adds_routecollections_from_constructor(self):
        r1 = Route("/", GET, dummyview, name="1")
        r2 = Route("/", POST, dummyview, name="2")
        r3 = Route("/", POST, dummyview, name="3")
        rc = RouteCollection([r1, RouteCollection([r2, r3])])
        assert [r.name for r in rc] == ["1", "2", "3"]

    def test_it_adds_dunderroutes_from_constructor(self):
        r1 = Route("/", GET, dummyview, name="1")
        r2 = Route("/", POST, dummyview, name="2")
        r3 = Route("/", POST, dummyview, name="3")

        class A:
            __routes__ = [r2, r3]

        rc = RouteCollection([r1, A()])
        assert [r.name for r in rc] == ["1", "2", "3"]

    def test_get_routes_matches_on_method(self):
        r_get = Route("/", GET, dummyview)
        r_post = Route("/", POST, dummyview)

        rc = RouteCollection([r_post, r_get])

        assert [r.route for r in rc.get_route_traversals("/", GET)] == [r_get]
        assert [r.route for r in rc.get_route_traversals("/", POST)] == [r_post]

    def test_get_routes_matches_on_path(self):
        r1 = Route("/1", GET, dummyview)
        r2 = Route("/2", GET, dummyview)

        rc = RouteCollection([r1, r2])

        assert [r.route for r in rc.get_route_traversals("/1", GET)] == [r1]
        assert [r.route for r in rc.get_route_traversals("/2", GET)] == [r2]

    def test_get_routes_can_match_all_methods(self):
        r1 = Route("/1", GET, dummyview)
        r2 = Route("/1", POST, dummyview)

        rc = RouteCollection([r1, r2])
        assert [r.route for r in rc.get_route_traversals("/1", None)] == [
            r1,
            r2,
        ]

    def test_route_returns_traversal_information_on_nested_routes(self):
        a = RouteCollection()
        b = RouteCollection()

        a_route = Route("/harvey", GET, dummyview)
        b_route = Route("/harvey", GET, dummyview)

        a.add_route(a_route)
        b.add_route(b_route)

        a_delegate_route = DelegateRoute("/rabbit", b)
        b_delegate_route = DelegateRoute("/hole", a)

        a.add_route(a_delegate_route)
        b.add_route(b_delegate_route)

        r = next(a.get_route_traversals("/rabbit/hole/rabbit/harvey", "GET"))

        assert r.collections_traversed == [
            (a, "", a_delegate_route, (), {}, (), {}),
            (b, "/rabbit", b_delegate_route, (), {}, (), {}),
            (a, "/rabbit/hole", a_delegate_route, (), {}, (), {}),
            (b, "/rabbit/hole/rabbit", b_route, (), {}, (), {}),
        ]

    def test_pathfor_works_with_positional_args(self):
        view = Mock(return_value=Response())
        rc = RouteCollection([Route("/<:str>", GET, view)])
        assert rc.pathfor(view, "x") == "/x"

    def test_replace_raises_route_not_found(self):
        a = RouteCollection()
        view = Mock(return_value=Response())
        a.route("/harvey", GET, view, name="harvey")
        with pytest.raises(RouteNotFound):
            a.replace("rabbit", None)

    def test_replace_selects_routes_by_name(self):
        a = RouteCollection()
        oldroute = Route("/", GET, Mock(), name="harvey")
        newroute = Route("/", GET, Mock())
        a.add_route(oldroute)
        a.replace("harvey", newroute)
        assert a.__routes__ == [newroute]

    def test_replace_selects_routes_by_view(self):
        a = RouteCollection()
        view = Mock(return_value=Response())
        oldroute = Route("/", GET, view)
        newroute = Route("/", GET, Mock())
        a.add_route(oldroute)
        a.replace(view, newroute)
        assert a.__routes__ == [newroute]

    def test_can_add_a_list_to_a_routecollection(self):
        r1 = Route("/", GET, Mock())
        r2 = Route("/", GET, Mock())
        assert (RouteCollection([r1]) + [r2]).__routes__ == [r1, r2]

    def test_can_add_route_to_routecollection(self):
        r1 = Route("/", GET, Mock())
        r2 = Route("/", GET, Mock())
        assert (RouteCollection([r1]) + r2).__routes__ == [r1, r2]

    def test_can_add_routecollection_to_route(self):
        r1 = Route("/", GET, Mock())
        r2 = Route("/", GET, Mock())
        assert (r1 + RouteCollection([r2])).__routes__ == [r1, r2]

    def test_can_add_routecollections(self):
        r1 = Route("/", GET, Mock())
        r2 = Route("/", GET, Mock())
        assert (RouteCollection([r1]) + RouteCollection([r2])).__routes__ == [
            r1,
            r2,
        ]

    def test_routecollections_can_be_used_in_classes(self):
        class MyViews(object):
            __routes__ = RouteCollection([Route("/", GET, "view")])

            def view(self):
                return Response()

        v = MyViews()
        app = FrescoApp()
        app.include("/", v)
        assert [r.route.getview(GET) for r in app.get_route_traversals("/", GET)] == [
            v.view
        ]

    def test_routecollections_in_classes_can_be_manipulated(self):
        class MyViews(object):
            __routes__ = RouteCollection([Route("/", GET, "view")])

            def view(self):
                return Response()

        class MyOtherViews(MyViews):
            __routes__ = copy(MyViews.__routes__)
            __routes__.replace("view", Route("/", GET, "another_view"))

            def another_view(self):
                return Response()

        v = MyOtherViews()
        app = FrescoApp()
        app.include("/", v)
        assert [r.route.getview(GET) for r in app.get_route_traversals("/", GET)] == [
            v.another_view
        ]

    def test_add_prefix_returns_prefixed_collection(self):
        rc = RouteCollection(
            [Route("/fish", GET, dummyview), Route("/beans", GET, dummyview)]
        )
        prefixed = rc.add_prefix("/jelly")
        assert [str(r.pattern) for r in prefixed] == [
            "/jelly/fish",
            "/jelly/beans",
        ]

    def test_it_binds_routes_to_an_instance(self):
        views = fixtures.CBV("test")
        rc = RouteCollection([views])
        view = next(rc.get_route_traversals("/", GET)).route.getview(GET)
        assert_method_bound_to(view, views)

    def test_it_binds_routes_to_an_instance_via_include(self):
        views = fixtures.CBV("test")
        rc = RouteCollection([])
        rc.include("/", views)
        view = next(rc.get_route_traversals("/", GET)).route.getview(GET)
        assert_method_bound_to(view, views)

    def test_including_twice_does_not_rebind_instance(self):
        views = fixtures.CBV("test")
        rc = RouteCollection([])
        rc.include("/", views)
        rc2 = RouteCollection([])
        rc2.include("/", rc)
        view = next(rc2.get_route_traversals("/", GET)).route.getview(GET)
        assert_method_bound_to(view, views)

    def test_it_binds_routes_via_a_string(self):
        rc = RouteCollection()
        rc.route("/", GET, "fresco.tests.fixtures.module_level_function")
        view = next(rc.get_route_traversals("/", GET)).route.getview(GET)
        assert view is fixtures.module_level_function

    def test_it_inserts_route(self):
        rc = RouteCollection([])
        a = Route("/", GET=lambda: None)
        b = Route("/", GET=lambda: None)
        rc.insert(0, a)
        rc.insert(0, b)
        assert list(rc) == [b, a]

    def test_len(self):
        assert len(RouteCollection()) == 0
        assert len(RouteCollection([Route("/", GET=lambda: None)])) == 1

    def test_getitem(self):
        a = Route("/", GET=lambda: None)
        b = Route("/", GET=lambda: None)
        c = Route("/", GET=lambda: None)
        rc = RouteCollection([a, b, c])
        assert rc[0] is a
        assert rc[1] is b
        assert rc[-1] is c
        assert rc[1:3].__routes__ == [b, c]

    def test_setitem(self):
        a = Route("/", GET=lambda: None)
        b = Route("/", GET=lambda: None)
        c = Route("/", GET=lambda: None)
        d = Route("/", GET=lambda: None)
        rc = RouteCollection([a, b])
        rc[0] = c
        assert rc.__routes__ == [c, b]
        rc[0:2] = [d]
        assert rc.__routes__ == [d]

    def test_it_routes_to_a_wsgi_app(self):
        def test_wsgi_app(
            mountpoint,
            path,
            expected_script_name,
            expected_path_info,
            rewrite_script_name=True,
        ):
            def wsgiapp(environ, start_response):
                wsgiapp.called = True  # type: ignore
                assert environ["SCRIPT_NAME"] == expected_script_name
                assert environ["PATH_INFO"] == expected_path_info
                start_response("204 no content", [])
                return []

            app = FrescoApp()
            app.route_wsgi(mountpoint, wsgiapp, rewrite_script_name=rewrite_script_name)
            with app.requestcontext(path):
                app.view()
            assert wsgiapp.called is True  # type: ignore

        test_wsgi_app("/prefix", "/prefix", "/prefix", "")
        test_wsgi_app("/prefix", "/prefix/", "/prefix", "/")
        test_wsgi_app("/", "/", "", "/")
        test_wsgi_app("/", "/foo", "", "/foo")
        test_wsgi_app("/prefix", "/prefix", "", "/prefix", rewrite_script_name=False)

    def test_it_routes_to_a_wsgi_app_by_dotted_path(self):
        app = FrescoApp()
        app.route_wsgi("/", "fresco.tests.fixtures.wsgi_app")
        with app.requestcontext("/"):
            response = app.view()
        assert list(response.content_iterator) == [b"ok"]

    def test_it_isolates_request_for_wsgi_app(self):
        params: t.Dict[str, t.Any] = {}
        outerapp = FrescoApp()
        innerapp = FrescoApp()
        innerapp.route("/y", GET=Response)
        outerapp.route_wsgi("/x", innerapp)

        innerapp.process_request(partial(params.__setitem__, "inner_request"))
        outerapp.process_request(partial(params.__setitem__, "outer_request"))

        with outerapp.requestcontext("/x/y"):
            outerapp.view()
        assert params["inner_request"] is not params["outer_request"]
        assert params["inner_request"].environ is not params["outer_request"].environ

    def test_fallthrough_can_be_applied_to_collection(self):
        a = Route("/fee", GET=lambda: Response(status=204))
        a = Route("/fie", GET=lambda: Response(status=204))
        b = Route("/foe", GET=lambda: Response(status=204))
        rc = RouteCollection([a, b])
        rc = rc.fallthrough_on(["204"])
        assert all(r.fallthrough_statuses == {204} for r in rc.__routes__)


class TestRoutefor(object):
    def test_routefor_with_view_function(self):
        def view():
            return Response(["ok"])

        app = FrescoApp()
        route = app.route("/foo", GET, view)

        with app.requestcontext():
            assert routefor(view) == route

    def test_routefor_with_string(self):
        app = FrescoApp()
        route = app.route("/myviewfunc", GET, fixtures.module_level_function)
        with app.requestcontext():
            assert routefor("fresco.tests.fixtures.module_level_function") == route

    def test_routefor_generates_first_route(self):
        def myviewfunc():
            return Response()

        app = FrescoApp()
        r1 = app.route("/1", GET, myviewfunc)
        app.route("/2", GET, myviewfunc)
        with app.requestcontext():
            assert routefor(myviewfunc) == r1


class TestDelegatedRoutes(object):
    def test_dispatch_to_delegated_route(self):
        def hello():
            return Response([b"hello"])

        inner = FrescoApp()
        inner.route("/hello", GET, hello)

        outer = FrescoApp()
        outer.delegate("/say", inner)

        with outer.requestcontext("/say/hello"):
            assert b"".join(outer.view().content) == b"hello"

    def test_can_delegate_to_plain_old_class(self):
        class Views(object):
            __routes__ = [Route("/hello", GET, "index")]

            def index(self):
                return Response(b"OK")

        app = FrescoApp()
        app.delegate("/foo", Views())

        with app.requestcontext("/foo/hello"):
            assert app.view().content == b"OK"

    def test_url_variables_are_passed(self):
        hello = Mock(return_value=Response())

        inner = FrescoApp()
        inner.route("/<i:str>", GET, hello)

        outer = FrescoApp()
        outer.delegate("/<o:str>", inner)

        with outer.requestcontext("/foo/bar"):
            outer.view()
            assert hello.call_args_list == [call(i="bar", o="foo")]

    def test_delegation_to_dynamic_routes(self):
        result = []

        class MyRoutes(object):
            __routes__ = [Route("/<inner:int>", GET, "view")]

            def __init__(self, **kwargs):
                self.kwargs = kwargs

            def view(self, **kwargs):
                result.append((self, kwargs))
                return Response()

        app = FrescoApp()
        app.delegate("/<outer:str>", MyRoutes, dynamic=True)
        with app.requestcontext("/one/2"):
            app.view()
            instance, inner_kwargs = result[0]
            assert instance.kwargs == {"outer": "one"}
            assert inner_kwargs == {"inner": 2}

    def test_dynamic_routes_are_never_shared(self):
        result = []

        class MyRoutes(object):
            __routes__ = [Route("", GET, "view")]

            def __init__(self, value):
                self.value = value

            def view(self):
                result.append(self.value)
                return Response()

        app = FrescoApp()
        app.delegate("/<value:str>", MyRoutes, dynamic=True)
        with app.requestcontext("/one"):
            app.view()
            v1 = result.pop()
        with app.requestcontext("/two"):
            app.view()
            v2 = result.pop()
        assert v1 == "one"
        assert v2 == "two", v2

    def test_pathfor_with_delegated_route(self):
        inner = FrescoApp()
        inner.route("/<i:str>", GET, lambda: None, name="inner-route")

        outer = FrescoApp()
        outer.delegate("/<o:str>", inner, name="delegation")

        with outer.requestcontext("/foo/bar"):
            assert outer.pathfor("delegation:inner-route", o="x", i="y") == "/x/y"

    def test_pathfor_with_dynamic_delegated_route(self):
        view = Mock(return_value=Response())

        def routecollectionfactory(*args, **kwargs):
            return RouteCollection([Route("/<i:str>", GET, view, name="inner-route")])

        rc = RouteCollection()
        rc.delegate("/<o:str>", routecollectionfactory, name="delegation", dynamic=True)

        assert rc.pathfor("delegation:inner-route", o="x", i="y") == "/x/y"

    def test_pathfor_with_dynamic_delegated_route_uses_default_args(self):
        view = Mock(return_value=Response())

        def routecollectionfactory(factoryarg1, factoryarg2):
            return RouteCollection([Route("/<i:str>", GET, view, name="inner-route")])

        rc = RouteCollection()
        rc.delegate(
            "/<factoryarg1:str>/<factoryarg2:str>",
            routecollectionfactory,
            factoryarg1_default="foo",
            factoryarg2_default=lambda r: "bar",
            name="delegation",
            dynamic=True,
        )

        assert rc.pathfor("delegation:inner-route", i="y") == "/foo/bar/y"

    def test_urlfor_with_dynamic_delegated_route_and_view_self(self):
        result = []

        class MyRoutes(object):
            __routes__ = [Route("/<inner:int>/view", GET, "view")]

            def __init__(self, **kwargs):
                self.kwargs = kwargs

            def view(self, **kwargs):
                result.append(urlfor(self.view, inner=3))
                return Response()

        app = FrescoApp()
        app.delegate("/<outer:str>", MyRoutes, dynamic=True)
        with app.requestcontext("/two/2/view"):
            app.view()
            assert result == ["http://localhost/two/3/view"]

    def test_urlgeneration_with_dynamic_routes(self):
        class Routable(object):
            __routes__ = [Route("/<b:int>", GET, "view", name="y")]

            def __init__(self, a):
                pass

            def view(self, b):
                return Response()

        app = FrescoApp()
        app.delegate("/<a:str>", Routable, dynamic=True, name="x")
        with app.requestcontext("/two/2/view"):
            assert urlfor("x:y", a="a", b=1) == "http://localhost/a/1"

    def test_delegated_routes_can_be_included(self):
        view = Mock(return_value=Response())

        inner = RouteCollection([Route("/baz", GET, view)])
        middle = RouteCollection([DelegateRoute("/bar", inner)])
        outer = FrescoApp()
        outer.include("/foo", middle)
        with outer.requestcontext("/foo/bar/baz"):
            outer.view()
            assert view.call_count == 1

    def test_not_found_is_returned(self):
        def inner():
            raise NotFound()

        outer = FrescoApp()
        outer.delegate("/foo", inner, dynamic=True)
        with outer.requestcontext("/foo/bar/baz"):
            response = outer.view()
            assert response.status_code == 404

    def test_not_found_causes_next_route_to_be_tried(self):
        def inner():
            raise NotFound()

        view = Mock(return_value=Response())

        outer = FrescoApp()
        outer.delegate("/foo", inner, dynamic=True)
        outer.route("/foo", GET, view)
        with outer.requestcontext("/foo"):
            outer.view()
            assert view.call_count == 1


class TestConverters(object):
    def test_str_converter_returns_unicode(self):
        from fresco.routing import StrConverter

        s = str("abc")
        assert isinstance(StrConverter().from_string(s), str)

    def test_register_coverter_acts_as_decorator(self):
        from fresco.routing import Converter, register_converter

        @register_converter("testconverter")
        class MyConverter(Converter):
            def from_string(self, s):
                return "bar"

        view = Mock(return_value=Response())

        app = FrescoApp()
        app.route("/<:testconverter>", GET, view)
        with app.requestcontext("/foo"):
            app.view()
            assert view.call_args == (("bar",), {}), view.call_args


class TestRRoute:
    def test_it_passes_the_request(self):
        from fresco.routing import RRoute

        view = Mock(return_value=Response())
        app = FrescoApp()
        app.add_route(RRoute("/", GET=view))
        with app.requestcontext("/") as c:
            app.view()
            assert view.call_args == ((c.request,), {})


class TestViewArgs(object):
    def test_it_uses_args(self):
        routes = RouteCollection([Route("/", GET, dummyview, args=(1, 2))])
        assert list(routes.get_route_traversals("/", GET)) == [
            tms.InstanceOf(RouteTraversal, args=(1, 2))
        ]

    def test_it_uses_view_args(self):
        routes = RouteCollection([Route("/", GET, dummyview, view_args=(1, 2))])
        assert list(routes.get_route_traversals("/", GET)) == [
            tms.InstanceOf(RouteTraversal, args=(1, 2))
        ]

    def test_it_appends_args_extracted_from_path(self):
        routes = RouteCollection([Route("/<:int>", GET, dummyview, view_args=(1, 2))])
        assert list(routes.get_route_traversals("/3", GET)) == [
            tms.InstanceOf(RouteTraversal, args=(1, 2, 3))
        ]

    def test_it_keeps_traversal_args_separate(self):
        routes = RouteCollection([Route("/<:int>", GET, dummyview, view_args=(1,))])
        assert list(routes.get_route_traversals("/2", GET)) == [
            tms.InstanceOf(
                RouteTraversal,
                args=(1, 2),
                collections_traversed=[
                    tms.InstanceOf(
                        TraversedCollection, args=(1, 2), traversal_args=(2,)
                    )
                ],
            )
        ]


class TestViewKwargs(object):
    def test_it_reads_from_route_kwargs(self):
        routes = RouteCollection([Route("/", GET, dummyview, x=1)])
        assert list(routes.get_route_traversals("/", GET)) == [
            tms.InstanceOf(RouteTraversal, kwargs={"x": 1})
        ]

    def test_it_reads_from_kwargs(self):
        routes = RouteCollection([Route("/", GET, dummyview, kwargs={"x": 1})])
        assert list(routes.get_route_traversals("/", GET)) == [
            tms.InstanceOf(RouteTraversal, kwargs={"x": 1})
        ]

    def test_it_reads_from_view_kwargs(self):
        routes = RouteCollection([Route("/", GET, dummyview, view_kwargs={"x": 1})])
        assert list(routes.get_route_traversals("/", GET)) == [
            tms.InstanceOf(RouteTraversal, kwargs={"x": 1})
        ]

    def test_it_keeps_traversal_kwargs_separate(self):
        routes = RouteCollection(
            [Route("/<x:int>", GET, dummyview, view_kwargs={"y": 1})]
        )
        assert list(routes.get_route_traversals("/2", GET)) == [
            tms.InstanceOf(
                RouteTraversal,
                kwargs={"x": 2, "y": 1},
                collections_traversed=[
                    tms.InstanceOf(
                        TraversedCollection,
                        kwargs={"y": 1, "x": 2},
                        traversal_kwargs={"x": 2},
                    )
                ],
            )
        ]


class TestRouteKwargs:
    def test_path_defaults_removed_from_view_kwargs(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route("/<test:int>", GET, view, name="test", test_default=1)
        with app.requestcontext("/2"):
            app.view()
            assert view.call_args_list == [call(test=2)]


class TestRouteClassIsPluggable(object):
    class CustomRoute(Route):
        pass

    def test_it_defaults_to_Route(self):
        routes = RouteCollection()
        assert routes.route_class is Route

    def test_it_accepts_route_class_arg(self):
        routes = RouteCollection(route_class=self.CustomRoute)
        assert routes.route_class is self.CustomRoute

    def test_it_uses_route_class_in_route_method(self):
        def myview():
            pass

        routes = RouteCollection(route_class=self.CustomRoute)
        routes.route("/", GET, myview)

        assert list(routes.get_route_traversals("/", GET)) == [
            tms.InstanceOf(RouteTraversal, route=tms.InstanceOf(self.CustomRoute))
        ]

    def test_it_uses_route_class_in_decorator(self):
        routes = RouteCollection(route_class=self.CustomRoute)

        @routes.route("/", GET)
        def myview():
            pass

        assert list(routes.get_route_traversals("/", GET)) == [
            tms.InstanceOf(RouteTraversal, route=tms.InstanceOf(self.CustomRoute))
        ]

    def test_custom_route_class_survives_include(self):
        routes = RouteCollection(route_class=self.CustomRoute)

        @routes.route("/", GET)
        def myview():
            pass

        routes2 = RouteCollection()
        routes2.include("/incl", routes)

        assert list(routes2.get_route_traversals("/incl/", GET)) == [
            tms.InstanceOf(RouteTraversal, route=tms.InstanceOf(self.CustomRoute))
        ]


class TestRouteTraversal(object):
    routes = RouteCollection()
    inner = RouteCollection()
    inner2 = RouteCollection()

    view = object()
    inner2.route("/<i:int>", GET=view, name="c")
    inner.delegate("/<i:int>", inner2, name="b")
    routes.delegate("/<i:int>", lambda i, r=inner: r, dynamic=True, name="a")

    def test_it_reconstructs_the_path(self):
        traversal = next(self.routes.get_route_traversals("/1/2/3", GET))

        # Do we get the original path back?
        assert traversal.build_path() == "/1/2/3"

    def test_it_modifies_the_path(self):
        traversal = next(self.routes.get_route_traversals("/1/2/3", GET))

        # Modify the traversal and check we get a modified path back
        assert traversal.replace("a", {"i": 0}).build_path() == "/0/2/3"

        assert traversal.replace("b", {"i": 0}).build_path() == "/1/0/3"
        assert traversal.replace("a:b", {"i": 0}).build_path() == "/1/0/3"

        assert traversal.replace("a:b:c", {"i": 0}).build_path() == "/1/2/0"
        assert traversal.replace("a:c", {"i": 0}).build_path() == "/1/2/0"
        assert traversal.replace("c", {"i": 0}).build_path() == "/1/2/0"
        assert traversal.replace(self.view, {"i": 0}).build_path() == "/1/2/0"

        # Passing no arguments should create a new traversal
        assert traversal.replace("a") is not traversal

        # ...identical to the original
        assert traversal.replace("a") == traversal

    def test_replace_raises_routenotfound(self):
        traversal = next(self.routes.get_route_traversals("/1/2/3", GET))
        with pytest.raises(RouteNotFound):
            traversal.replace("x")

        with pytest.raises(RouteNotFound):
            traversal.replace("a:x")

        with pytest.raises(RouteNotFound):
            traversal.replace("b:a")

        with pytest.raises(RouteNotFound):
            traversal.replace("a:b:c:d")

        with pytest.raises(RouteNotFound):
            traversal.replace("")


class TestRouteCache(object):
    def test_it_raises_exceptions(self):
        class CustomException(Exception):
            pass

        @register_converter("raises_exception")
        class RaisesException(Converter):
            """
            Convert latitude-longitude pairs into a ``LatLng`` named tuple.
            """

            pattern = r".*"

            def from_string(self, s):
                raise CustomException()

        rc = RouteCollection(
            [
                Route("/bacon", GET, Response),
                Route("/eggs/<eggs:raises_exception>", GET, Response),
            ],
            cache=True,
        )

        # Prime the cache
        list(rc.get_route_traversals("/bacon", GET))

        with pytest.raises(CustomException):
            list(rc.get_route_traversals("/eggs/x", GET))

        with pytest.raises(CustomException):
            list(rc.get_route_traversals("/eggs/x", GET))

        # Check a subsequent call to get_routes doesn't hang onto the exception
        list(rc.get_route_traversals("/bacon", GET))


class TestRouteAll:
    def test_route_all_matches_on_separator(self):
        def view():
            return Response()

        app = FrescoApp()
        app.route_all("/x", GET, view)

        assert len(list(app.get_route_traversals("/x", GET))) == 1
        assert len(list(app.get_route_traversals("/x/y", GET))) == 1
        assert len(list(app.get_route_traversals("/xy", GET))) == 0


class TestRequestParameter:
    def test_has_request_parameter(self):
        def a(request: Request):
            pass

        def b(request: t.Optional[Request]):
            pass

        def c(request: t.Union[Request, None]):
            pass

        def d(request: t.Union[None, Request]):
            pass

        def e(request: int):
            pass

        def f(request):
            pass

        assert _has_request_parameter(a) is True
        assert _has_request_parameter(b) is True
        assert _has_request_parameter(c) is True
        assert _has_request_parameter(d) is True
        assert _has_request_parameter(e) is False
        assert _has_request_parameter(f) is False

    def test_request_is_provided_automatically(self):
        def a(request: Request):
            return Response("a")

        def b():
            return Response("b")

        app = FrescoApp()
        app.route("/a", GET=a)
        app.route("/b", GET=b)

        with app.requestcontext("/a"):
            assert app.view().content == "a"

        with app.requestcontext("/b"):
            assert app.view().content == "b"



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from typing import Dict

from fresco import FrescoApp, Response
from fresco.util.wsgi import (
    ClosingIterator,
    StartResponseWrapper,
    getenv,
    setenv,
)


class Counter(object):
    value = 0

    def inc(self):
        self.value += 1


def start_response(status, headers, exc_info=None):
    pass


class _TestException(Exception):
    pass


class TestClosingIterator(object):
    def app(self, environ, start_response):
        start_response("200 OK", [("Content-Type: text/plain")])
        yield "Foo"
        yield "Bar"

    def test_close_called_after_iterator_finished(self):
        count = Counter()
        environ: Dict[str, str] = {}

        result = self.app(environ, start_response)
        result = ClosingIterator(result, count.inc)
        assert count.value == 0
        try:
            list(result)
        finally:
            result.close()
        assert count.value == 1

    def test_multiple_close_functions_called(self):
        count1 = Counter()
        count2 = Counter()
        environ: Dict[str, str] = {}

        result = self.app(environ, start_response)
        result = ClosingIterator(result, count1.inc, count2.inc)
        assert count1.value == 0
        assert count2.value == 0
        try:
            list(result)
        finally:
            result.close()
        assert count1.value == 1
        assert count2.value == 1


class TestStartResponseWrapper(object):
    def test_write(self):
        def wsgiapp(environ, start_response):
            start_response = StartResponseWrapper(start_response)
            write = start_response("200 OK", [("Content-Type", "text/plain")])

            write(b"cat")
            write(b"sat")
            write2 = start_response.call_start_response()
            write2(b"mat")
            return []

        with FrescoApp().requestcontext() as c:
            r = Response.from_wsgi(
                wsgiapp, c.request.environ, lambda status, headers: None
            )
            assert b"".join(r.content) == b"catsatmat"


class TestEnvironGetSet(object):
    wsgibytes = "ø".encode("UTF-8")
    wsgistr = wsgibytes.decode("ISO-8859-1")

    def test_getenv_returns_bytes(self):
        assert getenv({"x": self.wsgistr}, "x") == "ø".encode("UTF-8")

    def test_setenv_sets_str(self):
        env: Dict[str, str] = {}
        setenv(env, "x", "ø".encode("UTF-8"))
        assert env["x"] == self.wsgistr



from typing import Any
from typing import Callable
from typing import Dict
from typing import List
import sys


# Adapted from http://code.activestate.com/recipes/577969/
# Credit: Raymond Hettinger
# License: MIT
def cache_generator(original_function, maxsize):
    mapping: Dict[Any, Any] = {}
    mapping_get = mapping.get
    root: List[Any] = [None, None]
    root[:] = [root, root]
    value, exc_info = None, None
    size = 0

    PREV, NEXT = 0, 1
    while 1:
        key = yield value, exc_info
        link = mapping_get(key, root)
        exc_info = None
        if link is root:
            try:
                value, exc_info = original_function(key), None
            except Exception:
                exc_info = sys.exc_info()
                continue
            if size < maxsize:
                size += 1
            else:
                old_prev, old_next, old_key, old_value = root[NEXT]
                root[NEXT] = old_next
                old_next[PREV] = root
                del mapping[old_key]
            last = root[PREV]
            link = [last, root, key, value]
            mapping[key] = last[NEXT] = root[PREV] = link
        else:
            link_prev, link_next, key, value = link
            link_prev[NEXT] = link_next
            link_next[PREV] = link_prev
            last = root[PREV]
            last[NEXT] = root[PREV] = link
            link[PREV] = last
            link[NEXT] = root


def make_cache(original_function, maxsize=100) -> Callable[[Any], Any]:
    "Create a cache around a function that takes a single argument"
    c = cache_generator(original_function, maxsize)
    next(c)
    return c.send



# encoding=utf-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
Utilities for working with data on the HTTP level
"""

from binascii import hexlify
from collections import namedtuple
from collections import deque
from collections.abc import Collection
from collections.abc import Mapping
from email.header import Header
from email.message import Message
from email.parser import BytesFeedParser
from io import BytesIO
from itertools import chain
from typing import Dict
from typing import Iterator
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Callable
from typing import Union
from typing import Optional
import typing as t
from tempfile import SpooledTemporaryFile
import os
import re
import io

from urllib.parse import unquote_plus
from shutil import copyfileobj

import fresco
from fresco.defaults import DEFAULT_CHARSET
from fresco.exceptions import RequestParseError
from fresco.util.io import io_iterator
from fresco.util.io import ByteIterator
from fresco.util.wsgi import str_to_environ
from fresco.util.contentencodings import ALLOWED_ENCODINGS

KB = 1024
MB = 1024 * KB

#: Data chunk size to read from the input stream (wsgi.input)
CHUNK_SIZE = min(io.DEFAULT_BUFFER_SIZE, 1024)

ParsedContentType = namedtuple("ParsedContentType", "content_type encoding params")

ParsedField = Union["FileUpload", str]

token_pattern = r"[!#-\'*-.0-9A-Z\^-~]+"
quotedstringparts_pattern = r'(?:(\\.)|([^"\\]+))'
quotedstring_pattern = r'"(?:{})*"'.format(quotedstringparts_pattern)
quotedstring_parser = re.compile(r"{}".format(quotedstringparts_pattern))

parameter_parser = re.compile(
    r"\s*"
    r"(?P<name>{token})"
    r"\s*=\s*(?:({token})|({quotedstring}))\s*(?:;|$)".format(
        token=token_pattern, quotedstring=quotedstring_pattern
    )
)


def get_content_type_info(
    environ,
    default_type="application/octet-stream",
    default_encoding="iso-8859-1",
) -> ParsedContentType:
    """
    Read and parse the Content-Type header and return a
    :class:`ParsedContentType` object.
    """
    ct, params = parse_header(environ.get("CONTENT_TYPE", default_type))
    encoding = params.get("charset", default_encoding)
    if encoding is None or encoding.lower() not in ALLOWED_ENCODINGS:
        encoding = default_encoding
    return ParsedContentType(ct, encoding, params)


class TooBig(RequestParseError):
    """
    Request body is too big
    """

    def __init__(self, *args, **kwargs):
        super(TooBig, self).__init__(*args, **kwargs)
        self.response = fresco.Response.payload_too_large()


class MissingContentLength(RequestParseError):
    """
    No ``Content-Length`` header given
    """

    def __init__(self, *args, **kwargs):
        super(MissingContentLength, self).__init__(*args, **kwargs)
        self.response = fresco.Response.length_required()


def parse_parameters(s, preserve_backslashes=False) -> Dict[str, str]:
    """
    Return ``s`` parsed as a sequence of semi-colon delimited name=value pairs.

    Example usage::

        >>> from fresco.util.http import parse_parameters
        >>> parse_parameters('foo=bar')
        {'foo': 'bar'}
        >>> parse_parameters('foo="bar\\""')
        {'foo': 'bar"'}

    The ``preserve_backslashes`` flag is used to preserve IE compatibility
    for file upload paths, which it incorrectly encodes without escaping
    backslashes, eg::

        Content-Disposition: form-data; name="file"; filename="C:\\tmp\\Ext.js"

    (To be RFC compliant, the backslashes should be doubled up).
    """
    remaining = s.strip()
    if remaining == "":
        return {}

    params = {}
    while True:
        m = parameter_parser.match(remaining)
        if m is None:
            raise RequestParseError(
                "{!r}: expected parameter at character {}".format(
                    s, len(s) - len(remaining)
                ),
                content_type="text/plain",
            )
        groups = m.groups()
        name, value_token, value_qs = groups[:3]

        if value_token:
            params[name] = value_token
        else:
            if preserve_backslashes:
                params[name] = value_qs[1:-1]
            else:
                parts = quotedstring_parser.findall(value_qs)
                value = "".join((qp[1] if qp else t) for qp, t in parts)
                params[name] = value

        remaining = remaining[m.end() :]
        if not remaining:
            break

    return params


def parse_header(
    header: Union[str, Header],
    ie_workaround: bool = False,
    _broken_encoding_sniffer=re.compile(r'\\[^"\\]').search,
) -> Tuple[str, Dict[str, str]]:
    """
    Given a header, return a tuple of
    ``(value, {parameter_name: parameter_value}])``.

    Example usage::

        >>> parse_header("text/html; charset=UTF-8")
        ('text/html', {'charset': 'UTF-8'})
        >>> parse_header("multipart/form-data; boundary=-------7d91772e200be")
        ('multipart/form-data', {'boundary': '-------7d91772e200be'})
    """
    if isinstance(header, Header):
        # Python3's email.parser.Parser returns a Header object (rather than
        # a string) for values containing 8-bit characters. These are then
        # replaced by U+FFFD when converting the header to a string
        header = str(header)

    if ";" not in header:
        return header, {}

    preserve_backslashes = ie_workaround and _broken_encoding_sniffer(header)

    value, remaining = header.split(";", 1)
    return (
        value,
        parse_parameters(remaining.strip(), preserve_backslashes=preserve_backslashes),
    )


def parse_querystring(
    data: str,
    charset: Optional[str] = None,
    strict: bool = False,
    keep_blank_values: bool = True,
    unquote_plus=unquote_plus,
) -> List[Tuple[str, str]]:
    """
    Return ``(key, value)`` pairs from the given querystring::

        >>> list(parse_querystring('green%20eggs=ham;me=sam+i+am'))
        [('green eggs', 'ham'), ('me', 'sam i am')]

    :param data: The query string to parse.
    :param charset: Character encoding used to decode values. If not specified,
                    ``fresco.defaults.DEFAULT_CHARSET`` will be used.

    :param keep_blank_values: if True, keys without associated values will be
                              returned as empty strings. if False, no key,
                              value pair will be returned.

    :param strict: if ``True``, a ``RequestParseError`` will be raised on
                   parsing errors.
    """

    if charset is None:
        charset = DEFAULT_CHARSET

    result: List[Tuple[str, str]] = []
    append = result.append
    sep = "&" if "&" in data else ";"

    for item in data.split(sep):
        if not item:
            continue
        try:
            key, value = item.split("=", 1)
        except ValueError:
            if strict:
                raise RequestParseError(f"Bad query field: {item}")
            if not keep_blank_values:
                continue
            key, value = item, ""

        try:
            append((unquote_plus(key, charset), unquote_plus(value, charset)))
        except UnicodeDecodeError:
            raise RequestParseError(f"Invalid {charset} character data")
    return result


def parse_post(
    environ,
    _io: t.IO[bytes],
    default_charset: Optional[str] = None,
    max_size=16 * KB,
    max_multipart_size=2 * MB,
    ie_workaround=True,
) -> Tuple[Iterable[Tuple[str, ParsedField]], Optional[Callable]]:
    """\
    Parse the contents of an HTTP POST request, which may be either
    application/x-www-form-urlencoded or multipart/form-data encoded.

    Returned items are either tuples of (name, value) for simple string values
    or (name, FileUpload) for uploaded files.

    :param max_multipart_size: Maximum size of total data for a multipart form
                               submission

    :param max_size: The maximum size of data allowed to be read into memory.
                     For a application/x-www-form-urlencoded submission, this
                     is the maximum size of the entire data. For a
                     multipart/form-data submission, this is the maximum size
                     of any individual field (except file uploads).
    """
    ct, charset, ct_params = get_content_type_info(
        environ,
        "application/x-www-form-urlencoded",
        default_charset or DEFAULT_CHARSET,
    )

    try:
        content_length = int(environ["CONTENT_LENGTH"])
    except (TypeError, ValueError, KeyError):
        raise MissingContentLength()

    try:
        bytestream = io_iterator(_io, CHUNK_SIZE, maxlen=content_length)
        if ct == "application/x-www-form-urlencoded":
            if content_length > max_size:
                raise TooBig("Content Length exceeds permitted size")
            return (
                parse_querystring(b"".join(bytestream).decode("ASCII"), charset),
                None,
            )
        else:
            if content_length > max_multipart_size:
                raise TooBig("Content Length exceeds permitted size")
            try:
                boundary = ct_params["boundary"]
            except KeyError:
                raise RequestParseError(
                    "No boundary given in multipart/form-data content-type"
                )
            return parse_multipart(
                bytestream,
                boundary.encode("ASCII"),
                charset,
                max_size,
                ie_workaround=ie_workaround,
            )
    except UnicodeDecodeError:
        raise RequestParseError("Payload contains non ascii data")


class PostParser:
    close: t.Optional[Callable] = None

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        items, self.close = parse_post(*self.args, **self.kwargs)
        return items

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.close:
            self.close()


def get_body_bytes(environ, max_size=16 * KB) -> bytes:
    """
    Read a single message body from environ['wsgi.input'], returning a bytes
    object.
    """
    try:
        content_length = int(environ["CONTENT_LENGTH"])
    except (TypeError, ValueError, KeyError):
        raise MissingContentLength()

    if content_length > max_size:
        raise TooBig("Content Length exceeds permitted size")
    return b"".join(io_iterator(environ["wsgi.input"], maxlen=content_length))


class HTTPMessage(Message):
    """
    Represent HTTP request message headers
    """


def parse_multipart(
    stream: ByteIterator, boundary, default_charset, max_size, ie_workaround=True
) -> Tuple[Iterable[Tuple[str, ParsedField]], Optional[Callable]]:
    """
    Parse data encoded as ``multipart/form-data``.
    Return an iterator over tuples of ` (<field-name>, <data>)``, and an
    optional ``close`` function.

    ``data`` will be a string in the case of a regular input field, or a
    ``FileUpload`` instance if a file was uploaded.

    If a ``close`` function is returned, the caller must call it in order to
    close any temporary files created at the end of the request lifecycle.

    :param stream: input stream from which to read data
    :param boundary: multipart boundary string, as specified by the
                     ``Content-Disposition`` header
    :param default_charset: character set to use for encoding, if not specified
                            by a content-type header. In practice web browsers
                            don't supply a content-type header so this needs to
                            contain a sensible value.
    :param max_size: Maximum size in bytes for any non file upload part
    :param ie_workaround: If True (the default), enable a work around for IE's
                          broken content-disposition header encoding.
    """
    boundary_size = len(boundary)
    if boundary_size > 72:
        raise RequestParseError(
            "Malformed boundary string: must be no more than 70 characters, "
            "not counting the two leading hyphens (rfc 2046)"
        )

    assert (
        boundary_size + 4 < CHUNK_SIZE
    ), "CHUNK_SIZE cannot be smaller than the boundary string length + 4"

    peek = next(stream)
    while len(peek) < boundary_size + 4:
        peek += next(stream)

    if peek[0:2] != b"--":
        raise RequestParseError("Malformed POST data: expected two hypens")

    if peek[2 : boundary_size + 2] != boundary:
        raise RequestParseError("Malformed POST data: expected boundary")

    if peek[boundary_size + 2 : boundary_size + 4] != b"\r\n":
        raise RequestParseError("Malformed POST data: expected CRLF")

    stream = chain([peek[boundary_size + 4 :]], stream)
    open_files = set()
    fields = []

    try:
        while True:
            headers, data, stream = _read_multipart_field(stream, boundary, max_size)
            open_files.add(data)
            try:
                _, params = parse_header(
                    headers["Content-Disposition"], ie_workaround=ie_workaround
                )
            except KeyError:
                raise RequestParseError("Missing Content-Disposition header")

            try:
                name = params["name"]
            except KeyError:
                raise RequestParseError("Missing name in Content-Disposition header")

            is_file_upload = "Content-Type" in headers and "filename" in params
            if is_file_upload:
                data.seek(0)
                fu = FileUpload(params["filename"], headers, data)
                fields.append((name, fu))
            else:
                charset = parse_header(headers.get("Content-Type", ""))[1].get(
                    "charset", default_charset
                )
                if data.tell() > max_size:
                    data.close()
                    open_files.remove(data)
                    raise TooBig("Data block exceeds maximum permitted size")
                try:
                    data.seek(0)
                    fields.append((name, data.read().decode(charset)))
                    data.close()
                    open_files.remove(data)
                except UnicodeDecodeError:
                    raise RequestParseError(f"Invalid {charset} character data")

            peek = next(stream)
            if peek[:2] == b"\r\n":
                stream = chain([peek[2:]], stream)
            elif peek == b"--\r\n":
                if next(stream, None) is None:
                    break
                else:
                    RequestParseError("Boundary incorrectly terminated")
            else:
                raise RequestParseError("Boundary incorrectly terminated")
    except Exception:
        for f in open_files:
            f.close()
        raise

    close: Optional[Callable]  # type: ignore
    if open_files:

        def close():
            for f in open_files:
                f.close()

    else:
        close = None  # type: ignore

    return fields, close


def _read_multipart_field(
    stream: ByteIterator, boundary: bytes, max_size: int
) -> Tuple["HTTPMessage", SpooledTemporaryFile, ByteIterator]:
    """
    Read a single part from a multipart/form-data message and return a tuple of
    ``(headers, data, remainder)``.

    Iterator ``iostream`` must be positioned at the start of the header block
    for the field.

    The caller must call ``data.close()`` after consuming the data.

    ``headers`` is an instance of ``email.message.Message``.

    ``data`` is an instance of ``tempfile.SpooledTemporaryFile``.
    """
    output = SpooledTemporaryFile(max_size)
    parser = BytesFeedParser(_factory=HTTPMessage)
    parser._set_headersonly()  # type: ignore

    header_block, remainder, found = read_until(iter(stream), b"\r\n\r\n")

    deque(map(parser.feed, header_block), maxlen=1)
    if not found():
        raise RequestParseError("Incomplete data (expected header)")
    headers = parser.close()

    sep = b"\r\n--" + boundary
    data, remainder, found = read_until(remainder, sep)
    for chunk in data:
        output.write(chunk)

    # Fallen off the end of the input without having read a complete field?
    if not found():
        output.close()
        raise RequestParseError("Incomplete data (expected boundary)")
    output.flush()

    return headers, output, remainder


def read_until(
    stream: Iterator[bytes], delimiter: bytes
) -> Tuple[ByteIterator, ByteIterator, Callable[[], bool]]:
    """
    Return two iterators over byte stream `stream`` and a callable which
    indicates whether the delimiter was found.

    The first iterator yields all data up to ``delimiter``.
    The second iterator generates all remaining data.

    The first iterator must be exhausted before the second is iterated.
    The callable must only be called after the first iterator has been
    exhausted.
    """
    buf = b""
    found: t.Optional[bool] = None

    def _found():
        if found is None:
            raise AssertionError("The first iterator was not exhausted")
        return found

    def read_upto():
        nonlocal buf, found
        dlen = len(delimiter)
        for chunk in chain(stream, [b""]):
            buf += chunk
            is_at_end = chunk == b""
            if len(buf) > dlen + 4096 or is_at_end:
                before, sep, after = buf.partition(delimiter)
                if sep == b"":
                    if is_at_end:
                        found = False
                        yield buf
                        return
                    pos = len(buf) - dlen
                    if pos > 0:
                        yield buf[:pos]
                        buf = buf[pos:]
                else:
                    found = True
                    if before:
                        yield before
                    buf = after
                    return

    def remainder():
        if buf:
            yield buf
        yield from stream

    return read_upto(), remainder(), _found


class FileUpload(object):
    """\
    Represent a file uploaded in an HTTP form submission
    """

    def __init__(self, filename, headers, fileob):
        self.filename = filename
        self.headers = headers
        self.file = fileob

        # UNC/Windows path
        if self.filename[:2] == "\\\\" or self.filename[1:3] == ":\\":
            self.filename = self.filename[self.filename.rfind("\\") + 1 :]

    def save(self, fileob):
        """
        Save the upload to the file object or path ``fileob``

        :param fileob: a file-like object open for writing, or the path to the
                       file to be written
        """
        if isinstance(fileob, str):
            with open(fileob, "wb") as f:
                return self.save(f)

        self.file.seek(0)
        copyfileobj(self.file, fileob)


def encode_multipart(
    data: Optional[
        Union[
            Mapping[str, str],
            Collection[tuple[str, str]],
        ]
    ] = None,
    files: Optional[
        Collection[
            tuple[str, str, str, Union[bytes, t.Iterable[bytes], t.BinaryIO]]
        ]
    ] = None,
    charset="UTF-8",
    **kwargs
) -> tuple[bytes, dict[str, str]]:
    """
    Encode ``data`` using multipart/form-data encoding, returning a tuple
    of ``(<encoded data>, <environ items>)``.

    :param data: POST data to be encoded, either a dict or list of
                 ``(name, value)`` tuples.

    :param charset: Encoding used for any string values encountered in
                    ``data``

    :param files: collection of ``(name, filename, content_type, data)`` tuples.
                    ``data`` may be either a byte string, iterator or
                    file-like object.

    :param kwargs: other data items as keyword arguments
    :returns: a tuple of ``(<encoded_data>, <environ_items>)``,
              where ``encoded_data`` is a BytesIO object
              and ``environ`` is a dict containing the Content-Type and
              Content-Length headers encoded for inclusion in a WSGI environ
              dict.
    """

    if files is None:
        files = []

    data_items: Iterable[tuple[str, str]]
    if data is None:
        data_items = []
    elif isinstance(data, Mapping):
        data_items = iter(data.items())  # type: ignore
    else:
        data_items = data

    data_items = chain(data_items, kwargs.items())

    boundary = b"-------" + hexlify(os.urandom(16))
    alldata = chain(
        ((header_block(k), payload) for k, payload in data_items),
        ((file_header_block(k, fn, ct), payload) for k, fn, ct, payload in files),
    )

    CRLF = b"\r\n"
    post_data = BytesIO()
    post_data.write(b"--" + boundary)
    for headers, payload in alldata:
        post_data.write(CRLF)
        for name, value in headers:
            post_data.write("{0}: {1}\r\n".format(name, value).encode("ascii"))
        post_data.write(CRLF)
        write_payload(post_data, payload, charset)
        post_data.write(b"\r\n--" + boundary)
    post_data.write(b"--\r\n")
    length = post_data.tell()
    post_data.seek(0)
    wsgienv = {
        "CONTENT_LENGTH": str(length),
        "CONTENT_TYPE": str_to_environ(
            "multipart/form-data; boundary=" + boundary.decode("ascii")
        ),
    }

    return (post_data.getvalue(), wsgienv)


def header_block(name):
    return [("Content-Disposition", 'form-data; name="%s"' % (name,))]


def file_header_block(name, filename, content_type):
    return [
        (
            "Content-Disposition",
            'form-data; name="%s"; filename="%s"' % (name, filename),
        ),
        ("Content-Type", content_type),
    ]


def write_payload(stream, data, charset):
    "Write ``data`` to ``stream``, encoding as required"
    if hasattr(data, "read"):
        copyfileobj(data, stream)
    elif isinstance(data, bytes):
        stream.write(data)
    elif isinstance(data, str):
        stream.write(data.encode(charset))
    else:
        raise ValueError(data)



from fresco import FrescoApp
from fresco.routing import GET
from fresco.exceptions import BadRequest, RedirectTemporary


class TestResponseExceptions(object):
    def test_exception_is_converted_to_response(self):
        def redirector():
            raise RedirectTemporary("/foo")

        app = FrescoApp()
        app.route("/", GET, redirector)

        with app.requestcontext("/"):
            assert app.view().status_code == 302

    def test_exception_can_have_its_response_modified(self):
        def view():
            raise BadRequest(
                content="custom error message", content_type="application/foo"
            )

        app = FrescoApp()
        app.route("/", GET, view)

        with app.requestcontext("/"):
            response = app.view()
            assert response.status_code == 400
            assert "".join(response.content) == "custom error message"
            assert response.get_header("Content-Type") == "application/foo"



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
# See LICENSE.txt for terms of redistribution and use.

from io import BytesIO
from urllib.parse import quote
import re

from hypothesis import given, strategies as st
import pytest

from fresco import FrescoApp
from fresco import context
from fresco import Response
from fresco.exceptions import RequestParseError
from fresco.routing import POST
from fresco.util.http import FileUpload
from fresco.util.http import PostParser
from fresco.util.http import encode_multipart
from fresco.util.http import parse_parameters
from fresco.util.http import parse_post
from fresco.util.http import parse_querystring

from . import form_data


class TestParseQueryString(object):
    def p(self, value):
        return list(parse_querystring(value))

    def test_empty(self):
        assert self.p("") == []

    def test_simple_key_value(self):
        assert self.p("a=b") == [("a", "b")]

    def test_key_with_space(self):
        assert self.p("a+b=c") == [("a b", "c")]

    def test_value_with_space(self):
        assert self.p("a=b+c") == [("a", "b c")]

    def test_double_equals(self):
        assert self.p("a==b") == [("a", "=b")]

    def test_escaped_chars(self):
        assert self.p("%20==c%3D") == [(" ", "=c=")]

    def test_charset(self):
        assert self.p("a=el%20ni%C3%B1o") == [("a", "el niño")]


class TestParseMultipart(object):
    def _make_env(self, data: bytes):
        data = b"----XX\r\n" + data + b"----XX--\r\n"
        return (
            BytesIO(data),
            {
                "CONTENT_LENGTH": len(data),
                "CONTENT_TYPE": "multipart/form-data; boundary=--XX",
            },
        )

    def test_multipart(self):
        for data in form_data.multipart_samples:
            io = BytesIO(data["data"])  # type: ignore
            io.seek(0)
            environ = {
                "CONTENT_LENGTH": data["content_length"],
                "CONTENT_TYPE": data["content_type"],
            }
            with PostParser(environ, io, "UTF-8") as parsed:
                parsed = sorted(list(parsed))

                assert [name for name, value in parsed] == [
                    "empty-text-input",
                    "file-upload",
                    "text-input-ascii",
                    "text-input-unicode",
                ]

                assert parsed[0] == ("empty-text-input", "")
                assert parsed[2] == ("text-input-ascii", "abcdef")
                assert parsed[3] == (
                    "text-input-unicode",
                    b"\xce\xb1\xce\xb2\xce\xb3\xce\xb4".decode("utf8"),
                )

                fieldname, fileupload = parsed[1]
                assert isinstance(fileupload, FileUpload)
                assert fieldname == "file-upload"
                assert fileupload.filename == "test.data"
                assert fileupload.headers["content-type"] == "application/octet-stream"
                assert fileupload.file.read() == form_data.FILE_UPLOAD_DATA

    def test_non_ascii_headers(self):
        """
        Ensure that headers containing non-ascii characters are handled.
        NB the stdlib email.parser module that we use for parsing replaces
        any 8-bit characters with the unicode replacement character.
        """
        filename = "café.txt".encode("utf8")
        data = (
            b"----XXXXXX\r\n"
            b'Content-Disposition: form-data; name="u"; '
            b'filename="' + filename + b'"\r\n'
            b"Content-Type: application/octet-stream\r\n\r\n"
            b"1234567890\r\n"
            b"----XXXXXX--\r\n"
        )
        environ = {
            "CONTENT_LENGTH": len(data),
            "CONTENT_TYPE": "multipart/form-data; boundary=--XXXXXX",
        }
        f = BytesIO(data)
        with PostParser(environ, f, "UTF-8") as parsed:
            fieldname, fileupload = list(parsed)[0]
            assert isinstance(fileupload, FileUpload)
            assert fileupload.filename == "caf\uFFFD\uFFFD.txt"

    def test_malformed_headers_raise_exceptions(self):
        # Missing '=' in parameter
        data, env = self._make_env(
            b"Content-Disposition: form-data; name; \r\n\r\n" b"1234567890\r\n"
        )
        with pytest.raises(RequestParseError):
            list(parse_post(env, data))

    def test_fileupload_too_big(self):
        """\
        Verify that multipart/form-data encoded POST data raises an exception
        if the total data size exceeds request.MAX_SIZE bytes
        """

        def view():
            request = context.request
            request.MAX_MULTIPART_SIZE = 500
            request.get("f1")
            return Response(["ok"])

        app = FrescoApp()
        app.route("/", POST, view)

        with app.requestcontext_post(
            "/", files=[("f1", "filename.txt", "text/plain", "x" * 1000)]
        ):
            assert app.view().status == "413 Payload Too Large"

        with app.requestcontext_post(
            "/",
            files=[("f1", "filename.txt", "text/plain", "x" * 400)],
            data={"f2": "x" * 101},
        ):
            assert app.view().status == "413 Payload Too Large"

    def test_fileupload_with_invalid_content_length(self):
        def view():
            request = context.request
            request.get("f1")
            return Response(["ok"])

        app = FrescoApp()
        app.route("/", POST, view)

        with app.requestcontext_post(
            "/", files=[("f1", "filename.txt", "text/plain", "x" * 1000)]
        ) as c:
            c.request.environ["CONTENT_LENGTH"] = str("500")
            assert app.view().status == "400 Bad Request"

    def test_multipart_field_too_big(self):
        """
        Verify that multipart/form-data encoded POST data raises an exception
        if it contains a single field exceeding request.MAX_SIZE bytes
        """

        def view():
            request = context.request
            request.MAX_MULTIPART_SIZE = 500
            request.MAX_SIZE = 100
            request.get("f1")
            return Response(["ok"])

        app = FrescoApp()
        app.route("/", POST, view)

        with app.requestcontext_post("/", multipart=True, data=[("f1", "x" * 200)]):
            assert app.view().status == "413 Payload Too Large"


class TestParseFormEncodedData(object):
    char_latin1 = b"\xa3"
    char_utf8 = b"\xc2\xa3"
    char = char_latin1.decode("latin1")

    assert char_latin1.decode("latin1") == char_utf8.decode("utf8") == char

    def test_formencoded_data_too_big(self):
        """
        Verify that application/x-www-form-urlencoded POST data raises an
        exception if it exceeds request.MAX_SIZE bytes
        """

        def view():
            request = context.request
            request.MAX_SIZE = 100
            request.get("f1")
            return Response(["ok"])

        app = FrescoApp()
        app.route("/", POST, view)

        with app.requestcontext_post("/", data=[("f1", "x" * 200)]):
            assert app.view().status == "413 Payload Too Large"

    def test_posted_data_contains_non_ascii_chars(self):
        """
        Verify that it raises an exception if POST data contains invalid
        characters.
        """

        def view():
            context.request.get("f1")

        app = FrescoApp()
        app.route("/", POST, view)

        data = "foo=bár".encode("utf8")

        with app.requestcontext(
            "/",
            REQUEST_METHOD="POST",
            CONTENT_LENGTH=str(len(data)),
            CONTENT_TYPE="application/x-www-form-urlencoded",
            wsgi_input=data,
        ):
            response = app.view()
            assert response.status == "400 Bad Request"

    def test_non_utf8_data_posted(self):
        data = b"char=" + quote(self.char_latin1).encode("ascii")
        env = {
            "REQUEST_METHOD": "POST",
            "CONTENT_LENGTH": str(len(data)),
            "wsgi.input": BytesIO(data),
        }

        with FrescoApp().requestcontext(environ=env) as c:
            request = c.request
            request.charset = "latin1"
            assert request.form["char"] == self.char

    def test_non_utf8_data_getted(self):
        data = "char=" + quote(self.char_latin1)
        env = {"REQUEST_METHOD": "GET", "QUERY_STRING": data}

        with FrescoApp().requestcontext(environ=env) as c:
            request = c.request
            request.charset = "latin1"
            assert request.query["char"] == self.char


class TestEncodeMultipart(object):
    def test_it_encodes_a_data_dict(self):
        data, headers = encode_multipart([("foo", "bar baf")])
        assert b'Content-Disposition: form-data; name="foo"\r\n\r\nbar baf' in data

    def test_it_encodes_a_file_tuple(self):
        data, headers = encode_multipart(files=[("foo", "foo.txt", "ascii", "bar")])
        expected = (
            b"Content-Disposition: form-data; "
            b'name="foo"; filename="foo.txt"\r\n'
            b"Content-Type: ascii\r\n"
            b"\r\n"
            b"bar"
        )
        assert expected in data


class TestParseParameters(object):
    token = st.text(
        alphabet="".join(
            chr(c) for c in range(33, 127) if chr(c) not in '()<>@,;:\\/[]?="'
        ),
        min_size=1,
    )

    @given(items=st.lists(st.tuples(token, token)))
    def test_parse_tokens(self, items):
        s = ";".join("{}={}".format(k, v) for k, v in items)
        assert parse_parameters(s) == dict(items)

    @given(items=st.lists(st.tuples(token, st.text(min_size=1))))
    def test_parse_parameters(self, items):
        def escape(s):
            return re.sub(r'(["\\\r])', r"\\\1", s)

        s = ";".join('{}="{}"'.format(k, escape(v)) for k, v in items)
        assert parse_parameters(s) == dict(items)

    def test_semicolons_in_parameters(self):
        assert parse_parameters('name=";"') == {"name": ";"}

    def test_quoted_strings_in_parameters(self):
        assert parse_parameters('name="\\""') == {"name": '"'}

    def test_it_preserves_backslashes(self):
        assert parse_parameters('f="C:\\a.txt"', True) == {"f": "C:\\a.txt"}
